###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""DIRAC Production Management Class.

This class allows to monitor the progress of productions operationally.

Of particular use are the monitoring functions allowing to drill down
by site, minor status and application status for a given transformation.
"""
import os

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.Utilities.TimeUtilities import toString
from DIRAC.Core.Utilities.PromptUser import promptUser
from DIRAC.WorkloadManagementSystem.Client.JobMonitoringClient import JobMonitoringClient

from LHCbDIRAC.Interfaces.API.DiracLHCb import DiracLHCb
from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient


class DiracProduction(DiracLHCb):
    """class for managing productions."""

    def __init__(self, tsClientIn=None):
        """Instantiates the Workflow object and some default parameters."""

        super().__init__()

        if tsClientIn is None:
            self.transformationClient = TransformationClient()
        else:
            self.transformationClient = tsClientIn

        self.prodHeaders = {
            "AgentType": "SubmissionMode",
            "Status": "Status",
            "CreationDate": "Created",
            "TransformationName": "Name",
            "Type": "Type",
        }
        self.prodAdj = 22
        self.commands = {
            "start": ["Active", "Manual"],
            "stop": ["Stopped", "Manual"],
            "automatic": ["Active", "Automatic"],
            "manual": ["Active", "Manual"],
            "mctestmode": ["Testing", "Automatic"],
            "completed": ["Completed", "Manual"],
            "completing": ["Completing", "Automatic"],
            "cleaning": ["Cleaning", "Manual"],
            "flush": ["Flush", "Automatic"],
            "deleted": ["Deleted", "Manual"],
            "cleaned": ["Cleaned", "Manual"],
            "archived": ["Archived", "Manual"],
            "valinput": ["ValidatingInput", "Manual"],
            "valoutput": ["ValidatingOutput", "Manual"],
            "remove": ["RemovingFiles", "Manual"],
            "validated": ["ValidatedOutput", "Manual"],
            "removed": ["RemovedFiles", "Manual"],
        }

    def getProduction(self, productionID, printOutput=False):
        """Returns the metadata associated with a given production ID.

        Protects against
        LFN: being prepended and different types of production ID.
        """
        if not isinstance(productionID, ((int,), (str,))):
            return self._errorReport("Expected string, long or int for production ID")

        result = self.transformationClient.getTransformation(int(productionID))
        if not result["OK"]:
            return result

        # to fix TODO
        if printOutput:
            adj = self.prodAdj
            prodInfo = result["Value"]
            top = ""
            for i in self.prodHeaders.values():
                top += i.ljust(adj)
            message = ["ProductionID".ljust(adj) + top + "\n"]
            # very painful to make this consistent, better improved first on the server side
            productionID = str(productionID)
            info = (
                productionID.ljust(adj)
                + prodInfo["Status"].ljust(adj)
                + prodInfo["Type"].ljust(adj)
                + prodInfo["AgentType"].ljust(adj)
                + toString(prodInfo["CreationDate"]).ljust(adj)
                + prodInfo["TransformationName"].ljust(adj)
            )
            message.append(info)
            print("\n".join(message))
        return S_OK(result["Value"])

    def getProductionLoggingInfo(self, productionID, printOutput=False):
        """The logging information for the given production is returned.

        This includes the operation performed, any messages associated with
        the operation and the DN of the production manager performing it.
        """
        if not isinstance(productionID, ((int,), (str,))):
            return self._errorReport("Expected string, long or int for production ID")

        result = self.transformationClient.getTransformationLogging(int(productionID))
        if not result["OK"]:
            self.log.warn(f"Could not get transformation logging information for productionID {productionID}")
            return result
        if not result["Value"]:
            self.log.warn(f"No logging information found for productionID {productionID}")
            return result

        if not printOutput:
            return result

        infoM = (
            "ProdID".ljust(int(0.5 * self.prodAdj))
            + "Message".ljust(3 * self.prodAdj)
            + "DateTime [UTC]".ljust(self.prodAdj)
            + "AuthorCN".ljust(2 * self.prodAdj)
        )
        message = [infoM]
        for line in result["Value"]:
            infoL = (
                str(line["TransformationID"]).ljust(int(0.5 * self.prodAdj))
                + line["Message"].ljust(3 * self.prodAdj)
                + toString(line["MessageDate"]).ljust(self.prodAdj)
                + line["AuthorDN"].split("/")[-1].ljust(2 * self.prodAdj)
            )
            message.append(infoL)

        print("\nLogging summary for productionID " + str(productionID) + "\n\n" + "\n".join(message))

        return result

    def getProductionSummary(self, productionID=None, printOutput=False):
        """Returns a detailed summary for the productions in the system.

        If production ID is specified, the result is restricted to this
        value. If printOutput is specified, the result is printed to the
        screen.
        """
        if not isinstance(productionID, ((int,), (str,))):
            return self._errorReport("Expected string, long or int for production ID")

        result = self.transformationClient.getTransformationSummary()
        if not result["OK"]:
            return result

        if productionID:
            if int(productionID) in result["Value"]:
                newResult = S_OK()
                newResult["Value"] = {}
                newResult["Value"][int(productionID)] = result["Value"][int(productionID)]
                result = newResult
            else:
                self.log.info(
                    "Specified productionID was not found, \
          the list of active productions is:\n%s"
                    % ", ".join(str(pID) for pID in result["Value"])
                )
                return S_ERROR(f"Production ID {productionID} was not found")

        if printOutput:
            self._prettyPrint(result["Value"])

        return result

    def getProductionApplicationSummary(self, productionID, status=None, minorStatus=None, printOutput=False):
        """Returns an application status summary for the productions in the system.

        If printOutput is specified, the result is printed to the screen.
        This queries the WMS for the given productionID and provides an up-
        to-date snapshot of the application status combinations and
        associated WMS JobIDs.
        """
        if not isinstance(productionID, ((int,), (str,))):
            return self._errorReport("Expected string, long or int for production ID")

        statusDict = self.getProdJobMetadata(productionID, status, minorStatus)
        if not statusDict["OK"]:
            self.log.warn("Could not get production metadata information")
            return statusDict

        jobIDs = list(statusDict["Value"])
        if not jobIDs:
            return S_ERROR("No JobIDs with matching conditions found")

        self.log.verbose(f"Considering {len(jobIDs)} jobs with selected conditions")
        # now need to get the application status information
        result = JobMonitoringClient().getJobsApplicationStatus(jobIDs)
        if not result["OK"]:
            self.log.warn("Could not get application status for jobs list")
            return result

        appStatus = result["Value"]
        #    self._prettyPrint(appStatus)
        #    self._prettyPrint(statusDict['Value'])
        # Now format the result.
        summary = {}
        submittedJobs = 0
        doneJobs = 0
        for job, atts in statusDict["Value"].items():
            for key, val in atts.items():
                if key == "Status":
                    uniqueStatus = val.capitalize()
                    if uniqueStatus not in summary:
                        summary[uniqueStatus] = {}
                    if atts["MinorStatus"] not in summary[uniqueStatus]:
                        summary[uniqueStatus][atts["MinorStatus"]] = {}
                    if appStatus[job]["ApplicationStatus"] not in summary[uniqueStatus][atts["MinorStatus"]]:
                        summary[uniqueStatus][atts["MinorStatus"]][appStatus[job]["ApplicationStatus"]] = {}
                        summary[uniqueStatus][atts["MinorStatus"]][appStatus[job]["ApplicationStatus"]]["Total"] = 1
                        submittedJobs += 1
                        if uniqueStatus == "Done":
                            doneJobs += 1
                        summary[uniqueStatus][atts["MinorStatus"]][appStatus[job]["ApplicationStatus"]]["JobList"] = [
                            job
                        ]
                    else:
                        if appStatus[job]["ApplicationStatus"] not in summary[uniqueStatus][atts["MinorStatus"]]:
                            summary[uniqueStatus][atts["MinorStatus"]] = {}
                            summary[uniqueStatus][atts["MinorStatus"]][appStatus[job]["ApplicationStatus"]] = {}
                            summary[uniqueStatus][atts["MinorStatus"]][appStatus[job]["ApplicationStatus"]]["Total"] = 1
                            submittedJobs += 1
                            if uniqueStatus == "Done":
                                doneJobs += 1
                            summary[uniqueStatus][atts["MinorStatus"]][appStatus[job]["ApplicationStatus"]][
                                "JobList"
                            ] = [job]
                        else:
                            current = summary[uniqueStatus][atts["MinorStatus"]][appStatus[job]["ApplicationStatus"]][
                                "Total"
                            ]
                            summary[uniqueStatus][atts["MinorStatus"]][appStatus[job]["ApplicationStatus"]]["Total"] = (
                                current + 1
                            )
                            submittedJobs += 1
                            if uniqueStatus == "Done":
                                doneJobs += 1
                            jobList = summary[uniqueStatus][atts["MinorStatus"]][appStatus[job]["ApplicationStatus"]][
                                "JobList"
                            ]
                            jobList.append(job)
                            summary[uniqueStatus][atts["MinorStatus"]][appStatus[job]["ApplicationStatus"]][
                                "JobList"
                            ] = jobList

        if not printOutput:
            result = S_OK()
            if not status and not minorStatus:
                result["Totals"] = {"Submitted": int(submittedJobs), "Done": int(doneJobs)}
            result["Value"] = summary
            return result

        # If a printed summary is requested
        statAdj = int(0.5 * self.prodAdj)
        mStatAdj = int(2.0 * self.prodAdj)
        totalAdj = int(0.5 * self.prodAdj)
        exAdj = int(0.5 * self.prodAdj)
        message = f"\nJob Summary for ProductionID {productionID} considering status {status}"
        if minorStatus:
            message += f"and MinorStatus = {minorStatus}"

        message += ":\n\n"
        message += (
            "Status".ljust(statAdj)
            + "MinorStatus".ljust(mStatAdj)
            + "ApplicationStatus".ljust(mStatAdj)
            + "Total".ljust(totalAdj)
            + "Example".ljust(exAdj)
            + "\n"
        )
        for stat, metadata in summary.items():
            message += "\n"
            for minor, appInfo in metadata.items():
                message += "\n"
                for appStat, jobInfo in appInfo.items():
                    message += (
                        stat.ljust(statAdj)
                        + minor.ljust(mStatAdj)
                        + appStat.ljust(mStatAdj)
                        + str(jobInfo["Total"]).ljust(totalAdj)
                        + str(jobInfo["JobList"][0]).ljust(exAdj)
                        + "\n"
                    )

        # self._prettyPrint(summary)
        if status or minorStatus:
            return S_OK(summary)

        result = self.getProductionProgress(productionID)
        if not result["OK"]:
            self.log.warn("Could not get production progress information")
            return result

        if "Created" in result["Value"]:
            createdJobs = int(result["Value"]["Created"]) + submittedJobs
        else:
            createdJobs = submittedJobs

        percSub = int(100 * submittedJobs / createdJobs)
        percDone = int(100 * doneJobs / createdJobs)
        print(f"\nCurrent status of production {productionID}:\n")
        print(
            "Submitted".ljust(12)
            + str(percSub).ljust(3)
            + "%  ( "
            + str(submittedJobs).ljust(7)
            + "Submitted / ".ljust(15)
            + str(createdJobs).ljust(7)
            + " Created jobs )"
        )
        print(
            "Done".ljust(12)
            + str(percDone).ljust(3)
            + "%  ( "
            + str(doneJobs).ljust(7)
            + "Done / ".ljust(15)
            + str(createdJobs).ljust(7)
            + " Created jobs )"
        )
        result = S_OK()
        result["Totals"] = {"Submitted": int(submittedJobs), "Created": int(createdJobs), "Done": int(doneJobs)}
        result["Value"] = summary
        # self.pPrint(result)
        return result

    def getProductionJobSummary(self, productionID, status=None, minorStatus=None, printOutput=False):
        """Returns a job summary for the productions in the system.

        If printOutput is specified, the result is printed to the screen.
        This queries the WMS for the given productionID and provides an up-
        to-date snapshot of the job status combinations and associated WMS
        JobIDs.
        """
        if not isinstance(productionID, ((int,), (str,))):
            return self._errorReport("Expected string, long or int for production ID")

        statusDict = self.getProdJobMetadata(productionID, status, minorStatus)
        if not statusDict["OK"]:
            self.log.warn("Could not get production metadata information")
            return statusDict

        # Now format the result.
        summary = {}
        submittedJobs = 0
        doneJobs = 0
        for job, atts in statusDict["Value"].ietritems():
            for key, val in atts.items():
                if key == "Status":
                    uniqueStatus = val.capitalize()
                    if uniqueStatus not in summary:
                        summary[uniqueStatus] = {}
                    if atts["MinorStatus"] not in summary[uniqueStatus]:
                        summary[uniqueStatus][atts["MinorStatus"]] = {}
                        summary[uniqueStatus][atts["MinorStatus"]]["Total"] = 1
                        submittedJobs += 1
                        if uniqueStatus == "Done":
                            doneJobs += 1
                        summary[uniqueStatus][atts["MinorStatus"]]["JobList"] = [job]
                    else:
                        current = summary[uniqueStatus][atts["MinorStatus"]]["Total"]
                        summary[uniqueStatus][atts["MinorStatus"]]["Total"] = current + 1
                        submittedJobs += 1
                        if uniqueStatus == "Done":
                            doneJobs += 1
                        jobList = summary[uniqueStatus][atts["MinorStatus"]]["JobList"]
                        jobList.append(job)
                        summary[uniqueStatus][atts["MinorStatus"]]["JobList"] = jobList

        if not printOutput:
            result = S_OK()
            if not status and not minorStatus:
                result["Totals"] = {"Submitted": int(submittedJobs), "Done": int(doneJobs)}
            result["Value"] = summary
            return result

        # If a printed summary is requested
        statAdj = int(0.5 * self.prodAdj)
        mStatAdj = int(2.0 * self.prodAdj)
        totalAdj = int(0.5 * self.prodAdj)
        exAdj = int(0.5 * self.prodAdj)
        message = f"\nJob Summary for ProductionID {productionID} considering"
        if status:
            message += f" Status = {status}"
        if minorStatus:
            message += f" MinorStatus = {minorStatus}"
        if not status and not minorStatus:
            message += " all status combinations"

        message += ":\n\n"
        message += (
            "Status".ljust(statAdj)
            + "MinorStatus".ljust(mStatAdj)
            + "Total".ljust(totalAdj)
            + "Example".ljust(exAdj)
            + "\n"
        )
        for stat, metadata in summary.items():
            message += "\n"
            for minor, jobInfo in metadata.items():
                message += (
                    stat.ljust(statAdj)
                    + minor.ljust(mStatAdj)
                    + str(jobInfo["Total"]).ljust(totalAdj)
                    + str(jobInfo["JobList"][0]).ljust(exAdj)
                    + "\n"
                )

        print(message)
        # self._prettyPrint(summary)
        if status or minorStatus:
            return S_OK(summary)

        result = self.getProductionProgress(productionID)
        if not result["OK"]:
            return result

        if "Created" in result["Value"]:
            createdJobs = int(result["Value"]["Created"]) + submittedJobs
        else:
            createdJobs = submittedJobs

        percSub = int(100 * submittedJobs / createdJobs)
        percDone = int(100 * doneJobs / createdJobs)
        print(f"\nCurrent status of production {productionID}:\n")
        print(
            "Submitted".ljust(12)
            + str(percSub).ljust(3)
            + "%  ( "
            + str(submittedJobs).ljust(7)
            + "Submitted / ".ljust(15)
            + str(createdJobs).ljust(7)
            + " Created jobs )"
        )
        print(
            "Done".ljust(12)
            + str(percDone).ljust(3)
            + "%  ( "
            + str(doneJobs).ljust(7)
            + "Done / ".ljust(15)
            + str(createdJobs).ljust(7)
            + " Created jobs )"
        )
        result = S_OK()
        result["Totals"] = {"Submitted": int(submittedJobs), "Created": int(createdJobs), "Done": int(doneJobs)}
        result["Value"] = summary
        return result

    def getProductionSiteSummary(self, productionID, site=None, printOutput=False):
        """Returns a site summary for the productions in the system.

        If printOutput is specified, the result is printed to the screen.
        This queries the WMS for the given productionID and provides an up-
        to-date snapshot of the sites that jobs were submitted to.
        """
        if not isinstance(productionID, ((int,), (str,))):
            return self._errorReport("Expected string, long or int for production ID")

        statusDict = self.getProdJobMetadata(productionID, None, None, site)
        if not statusDict["OK"]:
            self.log.warn("Could not get production metadata information")
            return statusDict

        summary = {}
        submittedJobs = 0
        doneJobs = 0

        for job, atts in statusDict["Value"].items():
            for key, val in atts.items():
                if key == "Site":
                    uniqueSite = val
                    currentStatus = atts["Status"].capitalize()
                    if uniqueSite not in summary:
                        summary[uniqueSite] = {}
                    if currentStatus not in summary[uniqueSite]:
                        summary[uniqueSite][currentStatus] = {}
                        summary[uniqueSite][currentStatus]["Total"] = 1
                        submittedJobs += 1
                        if currentStatus == "Done":
                            doneJobs += 1
                        summary[uniqueSite][currentStatus]["JobList"] = [job]
                    else:
                        current = summary[uniqueSite][currentStatus]["Total"]
                        summary[uniqueSite][currentStatus]["Total"] = current + 1
                        submittedJobs += 1
                        if currentStatus == "Done":
                            doneJobs += 1
                        jobList = summary[uniqueSite][currentStatus]["JobList"]
                        jobList.append(job)
                        summary[uniqueSite][currentStatus]["JobList"] = jobList

        if not printOutput:
            result = S_OK()
            if not site:
                result = self.getProductionProgress(productionID)
                if not result["OK"]:
                    return result
                if "Created" in result["Value"]:
                    createdJobs = result["Value"]["Created"]
                result["Totals"] = {"Submitted": int(submittedJobs), "Done": int(doneJobs)}
            result["Value"] = summary
            return result

        # If a printed summary is requested
        siteAdj = int(1.0 * self.prodAdj)
        statAdj = int(0.5 * self.prodAdj)
        totalAdj = int(0.5 * self.prodAdj)
        exAdj = int(0.5 * self.prodAdj)
        message = f"\nSummary for ProductionID {productionID}"
        if site:
            message += f" at Site {site}"
        else:
            message += " at all Sites"
        message += ":\n\n"
        message += (
            "Site".ljust(siteAdj) + "Status".ljust(statAdj) + "Total".ljust(totalAdj) + "Example".ljust(exAdj) + "\n"
        )
        for siteStr, metadata in summary.items():
            message += "\n"
            for stat, jobInfo in metadata.items():
                message += (
                    siteStr.ljust(siteAdj)
                    + stat.ljust(statAdj)
                    + str(jobInfo["Total"]).ljust(totalAdj)
                    + str(jobInfo["JobList"][0]).ljust(exAdj)
                    + "\n"
                )

        print(message)
        # self._prettyPrint(summary)
        result = self.getProductionProgress(productionID)

        if not result["OK"]:
            return result

        if "Created" in result["Value"]:
            createdJobs = int(result["Value"]["Created"]) + submittedJobs
        else:
            createdJobs = submittedJobs

        percSub = int(100 * submittedJobs / createdJobs)
        percDone = int(100 * doneJobs / createdJobs)
        if not site:
            print(f"\nCurrent status of production {productionID}:\n")
            print(
                "Submitted".ljust(12)
                + str(percSub).ljust(3)
                + "%  ( "
                + str(submittedJobs).ljust(7)
                + "Submitted / ".ljust(15)
                + str(createdJobs).ljust(7)
                + " Created jobs )"
            )
            print(
                "Done".ljust(12)
                + str(percDone).ljust(3)
                + "%  ( "
                + str(doneJobs).ljust(7)
                + "Done / ".ljust(15)
                + str(createdJobs).ljust(7)
                + " Created jobs )"
            )
        result = S_OK()
        result["Totals"] = {"Submitted": int(submittedJobs), "Created": int(createdJobs), "Done": int(doneJobs)}
        result["Value"] = summary
        return result

    def getProductionProgress(self, productionID=None, printOutput=False):
        """Returns the status of jobs as seen by the production management
        infrastructure."""
        if not isinstance(productionID, ((int,), (str,))):
            return self._errorReport("Expected string, long or int for production ID")

        productionID = int(productionID)

        if not productionID:
            result = self._getActiveProductions()
            if not result["OK"]:
                return result
            productionID = result["Value"]
        else:
            productionID = [productionID]

        productionID = [str(x) for x in productionID]
        self.log.verbose(f"Will check progress for production(s):\n{', '.join(productionID)}")
        progress = {}
        for prod in productionID:
            # self._prettyPrint(result)
            result = self.transformationClient.getTransformationTaskStats(int(prod))
            if not result["Value"]:
                self.log.error(result)
                return result
            progress[int(prod)] = result["Value"]

        if not printOutput:
            return result
        idAdj = int(self.prodAdj)
        statAdj = int(self.prodAdj)
        countAdj = int(self.prodAdj)
        message = "ProductionID".ljust(idAdj) + "Status".ljust(statAdj) + "Count".ljust(countAdj) + "\n\n"
        for prod, info in progress.items():
            for status, count in info.items():
                message += str(prod).ljust(idAdj) + status.ljust(statAdj) + str(count).ljust(countAdj) + "\n"
            message += "\n"

        print(message)
        return result

    def _getActiveProductions(self, printOutput=False):
        """Returns a dictionary of active production IDs and their status, e.g.
        automatic, manual."""
        result = self.transformationClient.getTransformations()
        if not result["OK"]:
            return result
        prodList = result["Value"]
        currentProductions = {}
        for prodDict in prodList:
            self.log.debug(prodDict)
            if "AgentType" in prodDict and "TransformationID" in prodDict:
                prodID = prodDict["TransformationID"]
                status = prodDict["AgentType"]
                currentProductions[prodID] = status
                if status.lower() == "automatic":
                    self.log.verbose(f"Found active production {prodID} eligible to submit jobs")

        if printOutput:
            self._prettyPrint(currentProductions)

        return S_OK(currentProductions)

    def getProductionCommands(self):
        """Returns the list of possible commands and their meaning."""
        prodCommands = {}
        for keyword, statusSubMode in self.commands.items():
            prodCommands[keyword] = {"Status": statusSubMode[0], "SubmissionMode": statusSubMode[1]}
        return S_OK(prodCommands)

    def production(self, productionID, command, disableCheck=True):
        """Allows basic production management by supporting the following commands:

        - start : set production status to Active, job submission possible
        - stop : set production status to Stopped, no job submissions
        - automatic: set production submission mode to Automatic, e.g. submission via Agent
        - manual: set produciton submission mode to manual, e.g. dirac-production-submit
        """
        commands = self.commands

        if not isinstance(productionID, ((int,), (str,))):
            return self._errorReport("Expected string, long or int for production ID")

        productionID = int(productionID)
        if not isinstance(command, str):
            return self._errorReport("Expected string, for command")
        if not command.lower() in commands:
            return self._errorReport(f"Expected one of: {', '.join(commands)} for command string")

        self.log.verbose(f'Requested to change production {productionID} with command "{command.lower().capitalize()}"')
        if not disableCheck:
            result = promptUser(
                f'Do you wish to change production {productionID} with command "{command.lower().capitalize()}"? '
            )
            if not result["OK"]:
                self.log.info("Action cancelled")
                return S_OK("Action cancelled")
            if result["Value"] != "y":
                self.log.info("Doing nothing")
                return S_OK("Doing nothing")

        actions = commands[command]
        self.log.info(
            "Setting production status to %s and submission mode to %s for productionID %s"
            % (actions[0], actions[1], productionID)
        )
        result = self.transformationClient.setTransformationParameter(int(productionID), "Status", actions[0])
        if not result["OK"]:
            self.log.warn(f"Problem updating transformation status with result:\n{result}")
            return result
        self.log.verbose(f"Setting transformation status to {actions[0]} successful")
        result = self.transformationClient.setTransformationParameter(int(productionID), "AgentType", actions[1])
        if not result["OK"]:
            self.log.warn(f"Problem updating transformation agent type with result:\n{result}")
            return result
        self.log.verbose(f"Setting transformation agent type to {actions[1]} successful")
        return S_OK(f"Production {productionID} status updated")

    def productionFileSummary(
        self, productionID, selectStatus=None, outputFile=None, orderOutput=True, printSummary=False, printOutput=False
    ):
        """Allows to investigate the input files for a given production
        transformation and provides summaries / selections based on the file status
        if desired."""
        adj = 18
        ordering = "TaskID"
        if not orderOutput:
            ordering = "LFN"
        fileSummary = self.transformationClient.getTransformationFiles(
            condDict={"TransformationID": int(productionID)}, orderAttribute=ordering
        )
        if not fileSummary["OK"]:
            return fileSummary

        toWrite = ""
        totalRecords = 0
        summary = {}
        selected = 0
        if fileSummary["OK"]:
            for lfnDict in fileSummary["Value"]:
                totalRecords += 1
                record = ""
                recordStatus = ""
                for n, v in lfnDict.items():
                    record += str(n) + " = " + str(v).ljust(adj) + " "
                    if n == "Status":
                        recordStatus = v
                        if selectStatus == recordStatus:
                            selected += 1
                        if v in summary:
                            new = summary[v] + 1
                            summary[v] = new
                        else:
                            summary[v] = 1

                if outputFile and selectStatus:
                    if selectStatus == recordStatus:
                        toWrite += record + "\n"
                        if printOutput:
                            print(record)
                elif outputFile:
                    toWrite += record + "\n"
                    if printOutput:
                        print(record)
                else:
                    if printOutput:
                        print(record)

        if printSummary:
            print(f"\nSummary for {totalRecords} files in production {productionID}\n")
            print("Status".ljust(adj) + " " + "Total".ljust(adj) + "Percentage".ljust(adj) + "\n")
            for n, v in summary.items():
                percentage = int(100 * int(v) / totalRecords)
                print(str(n).ljust(adj) + " " + str(v).ljust(adj) + " " + str(percentage).ljust(2) + " % ")
            print("\n")

        if selectStatus and not selected:
            return S_ERROR(f'No files were selected for production {productionID} and status "{selectStatus}"')
        elif selectStatus and selected:
            print(
                '%s / %s files (%s percent) were found for production %s in status "%s"'
                % (selected, totalRecords, int(100 * int(selected) / totalRecords), productionID, selectStatus)
            )

        if outputFile:
            if os.path.exists(outputFile):
                print(f"Requested output file {outputFile} already exists, please remove this file to continue")
                return fileSummary

            with open(outputFile, "w") as fopen:
                fopen.write(toWrite)
            if not selectStatus:
                print(f"Wrote {totalRecords} lines to file {outputFile}")
            else:
                print(f'Wrote {selected} lines to file {outputFile} for status "{selectStatus}"')

        return fileSummary

    def checkFilesStatus(self, lfns, productionID="", printOutput=False):
        """Checks the given LFN(s) status in the productionDB.

        All productions are considered by default but can restrict to
        productionID.
        """
        if not isinstance(productionID, ((int,), (str,))):
            return self._errorReport("Expected string, long or int for production ID")

        if isinstance(lfns, str):
            lfns = lfns.replace("LFN:", "")
        elif isinstance(lfns, list):
            try:
                lfns = [str(lfnName.replace("LFN:", "")) for lfnName in lfns]
            except Exception as x:
                return self._errorReport(str(x), "Expected strings for LFN(s)")
        else:
            return self._errorReport("Expected single string or list of strings for LFN(s)")

        fileStatus = self.transformationClient.getFileSummary(lfns, int(productionID))
        if printOutput:
            self._prettyPrint(fileStatus["Value"])
        return fileStatus

    def getWMSProdJobID(self, jobID, printOutput=False):
        """This method takes the DIRAC WMS JobID and returns the Production JobID
        information."""
        result = self.getJobAttributes(jobID)
        if not result["OK"]:
            return result
        if "JobName" not in result["Value"]:
            return S_ERROR("Could not establish ProductionID / ProductionJobID, missing JobName")

        wmsJobName = result["Value"]["JobName"]
        prodID = wmsJobName.split("_")[0]
        prodJobID = wmsJobName.split("_")[1]
        info = {"WMSJobID": jobID, "JobName": wmsJobName, "ProductionID": prodID, "JobID": prodJobID}
        if printOutput:
            self._prettyPrint(info)
        return S_OK(info)

    def getProdJobInfo(self, productionID, jobID, printOutput=False):
        """Retrieve production job information from Production Manager service."""
        res = self.transformationClient.getTransformationTasks(
            condDict={"TransformationID": productionID, "TaskID": jobID}, inputVector=True
        )
        if not res["OK"]:
            return res
        if not res["Value"]:
            return S_ERROR(f"Job {jobID} not found for production {productionID}")
        jobInfo = res["Value"][0]
        if printOutput:
            self._prettyPrint(jobInfo)
        return S_OK(jobInfo)

    def selectProductionJobs(
        self, productionID, status=None, minorStatus=None, applicationStatus=None, site=None, owner=None, date=None
    ):
        """Wraps around DIRAC API selectJobs().

        Arguments correspond to the web page selections. By default, the
        date is the creation date of the production.
        """
        if not date:
            self.log.verbose(f"No Date supplied, setting old date for production {productionID}")
            date = "2001-01-01"
        return self.selectJobs(status, minorStatus, applicationStatus, site, owner, str(productionID).zfill(8), date)

    def extendProduction(self, productionID, numberOfJobs, printOutput=False):
        """Extend Simulation type Production by number of jobs.

        Usage: extendProduction <ProductionNameOrID> nJobs
        """
        if not isinstance(productionID, ((int,), (str,))):
            return self._errorReport("Expected string, long or int for production ID")

        if isinstance(numberOfJobs, str):
            try:
                numberOfJobs = int(numberOfJobs)
            except Exception as x:
                return self._errorReport(str(x), "Expected integer or string for number of jobs to submit")

        result = self.transformationClient.extendTransformation(int(productionID), numberOfJobs)
        if not result["OK"]:
            return self._errorReport(result, f"Could not extend production {productionID} by {numberOfJobs} jobs")

        if printOutput:
            print(f"Extended production {productionID} by {numberOfJobs} jobs")

        return result

    def getProdJobMetadata(self, productionID, status=None, minorStatus=None, site=None):
        """Function to get the WMS job metadata for selected fields.

        Given a production ID will return the current WMS status information
        for all jobs in that production starting from the creation date.
        """
        result = self.transformationClient.getTransformationParameters(int(productionID), ["CreationDate"])
        if not result["OK"]:
            self.log.warn(f"Problem getting production metadata for ID {productionID}:\n{result}")
            return result

        creationDate = toString(result["Value"]).split()[0]
        result = self.selectProductionJobs(
            productionID, status=status, minorStatus=minorStatus, site=site, date=creationDate
        )
        if not result["OK"]:
            self.log.warn(f"Problem selecting production jobs for ID {productionID}:\n{result}")
            return result

        jobsList = result["Value"]
        return self.getJobStatus(jobsList)

    def launchProduction(self, prod, publishFlag, testFlag, requestID, extend=0, tracking=0, MCsimflag=False):
        """Given a production object (prod), launch it It returns the productionID
        created."""

        if publishFlag is False and testFlag:
            gLogger.info("Test prod will be launched locally")
            result = prod.runLocal()
            if result["OK"]:
                gLogger.info("Template finished successfully")
                return S_OK()
            gLogger.error("Launching production: something wrong with execution!")
            return S_ERROR("Something wrong with execution!")

        result = prod.create(publish=publishFlag, requestID=requestID, reqUsed=tracking)

        if not result["OK"]:
            gLogger.error(f"Error during prod creation:\n{result['Message']}\ncheck that the wkf name is unique.")
            return result

        if publishFlag:
            prodID = result["Value"]
            msg = f"Production {prodID} successfully created "

            if extend:
                self.extendProduction(prodID, extend, printOutput=True)
                msg += f", extended by {extend} jobs"
            if MCsimflag:
                self.production(prodID, "mctestmode")
                msg = msg + " and started in mctestmode."
            elif testFlag:
                self.production(prodID, "manual")
                msg = msg + " and started in manual mode."
            else:
                self.production(prodID, "automatic")
                msg = msg + " and started in automatic mode."
            gLogger.notice(msg)

        else:
            prodID = 1
            gLogger.notice(
                "Production creation completed but not published (publishFlag was %s). \
      Setting ID = %s (useless, just for the test)"
                % (publishFlag, prodID)
            )

        return S_OK(prodID)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Test_RSS_Policy_Configurations."""
import unittest

import LHCbDIRAC.ResourceStatusSystem.Policy.Configurations as moduleTested

################################################################################


class Configurations_TestCase(unittest.TestCase):
    def setUp(self):
        """Setup."""

        self.moduleTested = moduleTested

    def tearDown(self):
        """Tear down."""

        del self.moduleTested


################################################################################


class Configurations_Success(Configurations_TestCase):
    def test_PoliciesMetaLHCb(self):
        policies = self.moduleTested.POLICIESMETA_LHCB

        policyKeys = {"description", "module", "command", "args"}

        for policyName, policy in policies.items():
            print(policyName)
            self.assertEqual(policyKeys, set(policy))
            self.assertTrue(isinstance(policy["description"], str))
            self.assertTrue(isinstance(policy["module"], str))
            self.assertTrue(isinstance(policy["command"], (tuple, None)))
            self.assertTrue(isinstance(policy["args"], (dict, None)))


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(Configurations_TestCase)
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(Configurations_Success))
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""This is the Data Integrity Client which allows the simple reporting of
problematic file and replicas to the IntegrityDB and their status correctly
updated in the FileCatalog."""
import errno
import re


from DIRAC import S_OK, gLogger
from DIRAC.Core.Utilities.Adler import compareAdler
from DIRAC.Core.Utilities.DErrno import cmpError
from DIRAC.Core.Utilities.ReturnValues import returnSingleResult
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
from DIRAC.Resources.Storage.StorageElement import StorageElement
from DIRAC.DataManagementSystem.Client.DataIntegrityClient import DataIntegrityClient as DIRACDataIntegrityClient
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
from LHCbDIRAC.DataManagementSystem.Client.ConsistencyChecks import ConsistencyChecks


class DataIntegrityClient(DIRACDataIntegrityClient):
    def __init__(self):
        """Extending DIRAC's DIRACDataIntegrityClient init."""
        super().__init__()

        self.cc = ConsistencyChecks()

    ##########################################################################
    #
    # This section contains the specific methods for BK->FC checks
    #

    def productionToCatalog(self, productionID):
        """This obtains the file information from the BK and checks these files are
        present in the FC."""
        gLogger.info("-" * 40)
        gLogger.info("Performing the BK->FC check")
        gLogger.info("-" * 40)
        res = self.__getProductionFiles(productionID)
        if not res["OK"]:
            return res
        noReplicaFiles = res["Value"]["GotReplicaNo"]
        yesReplicaFiles = res["Value"]["GotReplicaYes"]
        # For the files marked as existing we perfom catalog check
        res = self.cc._getCatalogMetadata(yesReplicaFiles)
        if not res["OK"]:
            return res
        catalogMetadata, missingCatalogFiles, zeroSizeFiles = res["Value"]
        if missingCatalogFiles:
            self._reportProblematicFiles(missingCatalogFiles, "LFNCatalogMissing")
        if zeroSizeFiles:
            self._reportProblematicFiles(zeroSizeFiles, "LFNZeroSize")

        # Try and get the metadata for files that shouldn't exist in the catalog
        if noReplicaFiles:
            res = self.__checkCatalogForBKNoReplicas(noReplicaFiles)
            if not res["OK"]:
                return res
            catalogMetadata.update(res["Value"])
        # Get the replicas for the files found to exist in the catalog
        res = self.cc._getCatalogReplicas(list(catalogMetadata))
        if not res["OK"]:
            return res
        replicas, zeroReplicaFiles = res["Value"]
        if zeroReplicaFiles:
            self._reportProblematicFiles(zeroReplicaFiles, "LFNZeroReplicas")
        resDict = {"CatalogMetadata": catalogMetadata, "CatalogReplicas": replicas}
        return S_OK(resDict)

    def __checkCatalogForBKNoReplicas(self, lfns):
        """Checks the catalog existence for given files."""
        gLogger.info(f"Checking the catalog existence of {len(lfns)} files")

        res = self.fc.getFileMetadata(lfns)
        if not res["OK"]:
            gLogger.error("Failed to get catalog metadata", res["Message"])
            return res
        allMetadata = res["Value"]["Successful"]
        existingCatalogFiles = list(allMetadata)
        if existingCatalogFiles:
            self._reportProblematicFiles(existingCatalogFiles, "BKReplicaNo")
        gLogger.info("Checking the catalog existence of files complete")
        return S_OK(allMetadata)

    def __getProductionFiles(self, productionID):
        """This method queries the bookkeeping and obtains the file metadata for
        the given production."""
        gLogger.info(f"Attempting to get files for production {productionID}")
        res = BookkeepingClient().getProductionFiles(productionID, "ALL")
        if not res["OK"]:
            return res
        yesReplicaFiles = []
        noReplicaFiles = []
        badReplicaFiles = []
        badBKFileSize = []
        badBKGUID = []
        allMetadata = res["Value"]
        gLogger.info(f"Obtained at total of {len(allMetadata)} files")
        totalSize = 0
        for lfn, bkMetadata in allMetadata.items():
            if bkMetadata["FileType"] != "LOG":
                if bkMetadata["GotReplica"] == "Yes":
                    yesReplicaFiles.append(lfn)
                    if bkMetadata["FileSize"]:
                        totalSize += int(bkMetadata["FileSize"])
                elif bkMetadata["GotReplica"] == "No":
                    noReplicaFiles.append(lfn)
                else:
                    badReplicaFiles.append(lfn)
                if not bkMetadata["FileSize"]:
                    badBKFileSize.append(lfn)
                if not bkMetadata["GUID"]:
                    badBKGUID.append(lfn)
        if badReplicaFiles:
            self._reportProblematicFiles(badReplicaFiles, "BKReplicaBad")
        if badBKFileSize:
            self._reportProblematicFiles(badBKFileSize, "BKSizeBad")
        if badBKGUID:
            self._reportProblematicFiles(badBKGUID, "BKGUIDBad")
        gLogger.info(f"{len(yesReplicaFiles)} files marked with replicas with total size {totalSize} bytes")
        gLogger.info(f"{len(noReplicaFiles)} files marked without replicas")
        resDict = {"BKMetadata": allMetadata, "GotReplicaYes": yesReplicaFiles, "GotReplicaNo": noReplicaFiles}
        return S_OK(resDict)

    ##########################################################################
    #
    # This section contains the specific methods for FC->BK checks
    #

    def catalogDirectoryToBK(self, lfnDir):
        """This obtains the replica and metadata information from the catalog for
        the supplied directory and checks against the BK."""
        gLogger.info("-" * 40)
        gLogger.info("Performing the FC->BK check")
        gLogger.info("-" * 40)
        if isinstance(lfnDir, str):
            lfnDir = [lfnDir]
        res = self.__getCatalogDirectoryContents(lfnDir)
        if not res["OK"]:
            return res
        replicas = res["Value"]["Replicas"]
        catalogMetadata = res["Value"]["Metadata"]
        resDict = {"CatalogMetadata": catalogMetadata, "CatalogReplicas": replicas}
        if not catalogMetadata:
            gLogger.warn(f"No files found in directory {lfnDir}")
            return S_OK(resDict)
        lfns = []
        for repDict in replicas:
            lfns.append(repDict)
        missingLFNs, noFlagLFNs, _okLFNs = self.cc._getBKMetadata(lfns)
        if missingLFNs:
            self._reportProblematicFiles(missingLFNs, "LFNBKMissing")
        if noFlagLFNs:
            self._reportProblematicFiles(noFlagLFNs, "BKReplicaNo")
        return S_OK(resDict)

    def catalogFileToBK(self, lfns):
        """This obtains the replica and metadata information from the catalog and
        checks against the storage elements."""
        gLogger.info("-" * 40)
        gLogger.info("Performing the FC->BK check")
        gLogger.info("-" * 40)
        if isinstance(lfns, str):
            lfns = [lfns]

        res = self.cc._getCatalogMetadata(lfns)
        if not res["OK"]:
            return res
        catalogMetadata, missingCatalogFiles, zeroSizeFiles = res["Value"]
        if missingCatalogFiles:
            self._reportProblematicFiles(missingCatalogFiles, "LFNCatalogMissing")
        if zeroSizeFiles:
            self._reportProblematicFiles(zeroSizeFiles, "LFNZeroSize")

        res = self.cc._getCatalogReplicas(list(catalogMetadata))
        if not res["OK"]:
            return res
        replicas, _zeroReplicaFiles = res["Value"]

        lfns = []
        for repDict in replicas:
            lfns.append(repDict)
        missingLFNs, noFlagLFNs, _okLFNs = self.cc._getBKMetadata(lfns)
        if missingLFNs:
            self._reportProblematicFiles(missingLFNs, "LFNBKMissing")
        if noFlagLFNs:
            self._reportProblematicFiles(noFlagLFNs, "BKReplicaNo")

        resDict = {"CatalogMetadata": catalogMetadata, "CatalogReplicas": replicas}
        return S_OK(resDict)

    ##########################################################################
    #
    # This section contains the resolution methods for various prognoses
    #

    def resolveBKReplicaYes(self, problematicDict):
        """This takes the problematic dictionary returned by the integrity DB and
        resolved the BKReplicaYes prognosis."""
        lfn = problematicDict["LFN"]
        fileID = problematicDict["FileID"]

        res = returnSingleResult(self.fc.exists(lfn))
        if not res["OK"]:
            return self.__returnProblematicError(fileID, res)
        removeBKFile = False
        # If the file does not exist in the catalog
        if not res["Value"]:
            gLogger.info("BKReplicaYes file (%d) does not exist in the catalog. Removing..." % fileID)
            removeBKFile = True
        else:
            gLogger.info("BKReplicaYes file (%d) found to exist in the catalog" % fileID)
            # If the file has no replicas in the catalog
            res = returnSingleResult(self.fc.getReplicas(lfn))
            if (not res["OK"]) and (res["Message"] == "File has zero replicas"):
                gLogger.info("BKReplicaYes file (%d) found to exist without replicas. Removing..." % fileID)
                removeBKFile = True
        if removeBKFile:
            # Remove the file from the BK because it does not exist
            res = returnSingleResult(FileCatalog(catalogs=["BookkeepingDB"]).removeFile(lfn))
            if not res["OK"]:
                return self.__returnProblematicError(fileID, res)
            gLogger.info("BKReplicaYes file (%d) removed from bookkeeping" % fileID)
        return self.__updateCompletedFiles("BKReplicaYes", fileID)

    def resolveBKReplicaNo(self, problematicDict):
        """This takes the problematic dictionary returned by the integrity DB and
        resolved the BKReplicaNo prognosis."""
        lfn = problematicDict["LFN"]
        fileID = problematicDict["FileID"]

        res = returnSingleResult(self.fc.exists(lfn))
        if not res["OK"]:
            return self.__returnProblematicError(fileID, res)
        # If the file exists in the catalog
        if not res["Value"]:
            return self.__updateCompletedFiles("BKReplicaNo", fileID)
        gLogger.info("BKReplicaNo file (%d) found to exist in the catalog" % fileID)
        # and has available replicas
        res = returnSingleResult(self.fc.getCatalogReplicas(lfn))
        if not res["OK"]:
            return self.__returnProblematicError(fileID, res)
        if not res["Value"]:
            gLogger.info("BKReplicaNo file (%d) found to have no replicas" % fileID)
            return self.changeProblematicPrognosis(fileID, "LFNZeroReplicas")
        gLogger.info("BKReplicaNo file (%d) found to have replicas" % fileID)
        res = returnSingleResult(FileCatalog(catalogs=["BookkeepingDB"]).addFile(lfn))
        if not res["OK"]:
            return self.__returnProblematicError(fileID, res)
        return self.__updateCompletedFiles("BKReplicaNo", fileID)

    def checkPhysicalFiles(self, replicas, catalogMetadata, ses=[], fixIt=False):
        """This obtains takes the supplied replica and metadata information
        obtained from the catalog and checks against the storage elements."""
        gLogger.info("-" * 40)
        gLogger.info("Performing the FC->SE check")
        gLogger.info("-" * 40)

        seLfns = {}
        for lfn, replicaDict in replicas.items():
            for se in replicaDict:
                if (ses) and (se not in ses):
                    continue
                seLfns.setdefault(se, []).append(lfn)
        gLogger.info(f"{'Storage Element'.ljust(20)} {'Replicas'.rjust(20)}")

        for se in seLfns:
            res = StorageElement(se).isValid("Check")
            if not res["OK"]:
                gLogger.error("Unable to checkPhysicalFiles at", f"{se}: {res['Message']}")
                return res

        for se, lfns in sorted(seLfns.items()):
            sizeMismatch = []
            checksumMismatch = []
            checksumBadInFC = []
            res = self.__checkPhysicalFileMetadata(lfns, se)
            if not res["OK"]:
                gLogger.error("Failed to get physical file metadata.", res["Message"])
                return res
            for lfn, metadata in res["Value"].items():
                if lfn in catalogMetadata:
                    if (metadata["Size"] != catalogMetadata[lfn]["Size"]) and (metadata["Size"] != 0):
                        sizeMismatch.append((lfn, "deprecatedUrl", se, "CatalogPFNSizeMismatch"))
                    if not metadata["Checksum"]:
                        checksumMismatch.append(
                            (
                                lfn,
                                "deprecatedUrl",
                                se,
                                "Error fetching checksum",
                            )
                        )
                    elif not compareAdler(metadata["Checksum"], catalogMetadata[lfn]["Checksum"]):
                        if compareAdler(
                            metadata["Checksum"].replace("x", "0"), catalogMetadata[lfn]["Checksum"].replace("x", "0")
                        ):
                            checksumBadInFC.append(
                                (
                                    lfn,
                                    "deprecatedUrl",
                                    se,
                                    f"{metadata['Checksum']} {catalogMetadata[lfn]['Checksum']}",
                                )
                            )
                        else:
                            checksumMismatch.append(
                                (
                                    lfn,
                                    "deprecatedUrl",
                                    se,
                                    f"{metadata['Checksum']} {catalogMetadata[lfn]['Checksum']}",
                                )
                            )
            if sizeMismatch:
                self.reportProblematicReplicas(sizeMismatch, se, "CatalogPFNSizeMismatch", fixIt=fixIt)
            if checksumMismatch:
                self.reportProblematicReplicas(checksumMismatch, se, "CatalogChecksumMismatch", fixIt=fixIt)
            if checksumBadInFC:
                self.reportProblematicReplicas(checksumBadInFC, se, "CatalogChecksumToBeFixed", fixIt=fixIt)
        return S_OK()

    def __checkPhysicalFileMetadata(self, lfns, se):
        """Check obtain the physical file metadata and check the files are
        available."""
        gLogger.info(f"Checking the integrity of {len(lfns)} physical files at {se}")
        res = StorageElement(se).getFileMetadata(lfns)
        if not res["OK"]:
            gLogger.error("Failed to get metadata for lfns.", res["Message"])
            return res
        pfnMetadataDict = res["Value"]["Successful"]
        # If the replicas are completely missing
        missingReplicas = []
        for lfn, reason in res["Value"]["Failed"].items():
            if re.search("File does not exist", reason) or cmpError(reason, errno.ENOENT):
                missingReplicas.append((lfn, "deprecatedUrl", se, "PFNMissing"))
        if missingReplicas:
            self.reportProblematicReplicas(missingReplicas, se, "PFNMissing")
        lostReplicas = []
        unavailableReplicas = []
        zeroSizeReplicas = []
        # If the files are not accessible
        for lfn, metadata in pfnMetadataDict.items():
            if metadata.get("Lost", False):
                lostReplicas.append((lfn, se, "PFNLost"))
            if metadata.get("Unavailable", not metadata["Accessible"]):
                unavailableReplicas.append((lfn, "deprecatedUrl", se, "PFNUnavailable"))
            if metadata["Size"] == 0:
                zeroSizeReplicas.append((lfn, "deprecatedUrl", se, "PFNZeroSize"))
        if lostReplicas:
            self.reportProblematicReplicas(lostReplicas, se, "PFNLost")
        if unavailableReplicas:
            self.reportProblematicReplicas(unavailableReplicas, se, "PFNUnavailable")
        if zeroSizeReplicas:
            self.reportProblematicReplicas(zeroSizeReplicas, se, "PFNZeroSize")
        gLogger.info(f"Checking the integrity of physical files at {se} complete")
        return S_OK(pfnMetadataDict)

    def reportProblematicReplicas(self, replicaTuple, se, reason, fixIt=False):
        """Simple wrapper function around setReplicaProblematic."""
        gLogger.info(f"The following {len(replicaTuple)} files had {reason} at {se}")
        for lfn, pfn, se, reason1 in sorted(replicaTuple):
            if reason1 == reason:
                reason1 = ""
            if lfn:
                gLogger.info(lfn, reason1)
            else:
                gLogger.info(pfn, reason1)
        if fixIt:
            res = self.setReplicaProblematic(replicaTuple, sourceComponent="DataIntegrityClient")
            if not res["OK"]:
                gLogger.info("Failed to update integrity DB with replicas", res["Message"])
            else:
                gLogger.info("Successfully updated integrity DB with replicas")

    ##########################################################################
    #
    # This section contains the specific methods for obtaining replica and metadata information from the catalog
    #

    def __getCatalogDirectoryContents(self, lfnDir):
        """Obtain the contents of the supplied directory."""
        gLogger.info(f"Obtaining the catalog contents for {len(lfnDir)} directories")

        activeDirs = list(lfnDir)
        allFiles = {}
        while len(activeDirs) > 0:
            currentDir = activeDirs[0]
            res = self.fc.listDirectory(currentDir, verbose=True)
            activeDirs.remove(currentDir)
            if not res["OK"]:
                gLogger.error("Failed to get directory contents", res["Message"])
                return res
            elif currentDir in res["Value"]["Failed"]:
                gLogger.error("Failed to get directory contents", f"{currentDir} {res['Value']['Failed'][currentDir]}")
            else:
                dirContents = res["Value"]["Successful"][currentDir]
                activeDirs.extend(dirContents["SubDirs"])
                allFiles.update(dirContents["Files"])

        zeroReplicaFiles = []
        zeroSizeFiles = []
        allReplicaDict = {}
        allMetadataDict = {}
        for lfn, lfnDict in allFiles.items():
            lfnReplicas = {}
            for se, replicaDict in lfnDict["Replicas"].items():
                lfnReplicas[se] = replicaDict["PFN"]
            if not lfnReplicas:
                zeroReplicaFiles.append(lfn)
            allReplicaDict[lfn] = lfnReplicas
            allMetadataDict[lfn] = lfnDict["MetaData"]
            if lfnDict["MetaData"]["Size"] == 0:
                zeroSizeFiles.append(lfn)
        if zeroReplicaFiles:
            self._reportProblematicFiles(zeroReplicaFiles, "LFNZeroReplicas")
        if zeroSizeFiles:
            self._reportProblematicFiles(zeroSizeFiles, "LFNZeroSize")
        gLogger.info(f"Obtained at total of {len(allMetadataDict)} files for the supplied directories")
        resDict = {"Metadata": allMetadataDict, "Replicas": allReplicaDict}
        return S_OK(resDict)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
:mod:  WorkflowTaskAgent

.. module:  WorkflowTaskAgent

:synopsis:  Extension of the DIRAC WorkflowTaskAgent, to use LHCb clients.
"""

from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.TransformationSystem.Agent.WorkflowTaskAgent import WorkflowTaskAgent as DIRACWorkflowTaskAgent

from LHCbDIRAC.Interfaces.API.LHCbJob import LHCbJob
from LHCbDIRAC.TransformationSystem.Client.TaskManager import LHCbWorkflowTasks
from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient

AGENT_NAME = "Transformation/WorkflowTaskAgent"


class WorkflowTaskAgent(DIRACWorkflowTaskAgent):
    """An AgentModule class to submit workflow tasks."""

    def __init__(self, *args, **kwargs):
        """c'tor."""
        DIRACWorkflowTaskAgent.__init__(self, *args, **kwargs)

    def _getClients(self):
        """LHCb clients."""
        res = DIRACWorkflowTaskAgent._getClients(self)

        outputDataModule = Operations().getValue(
            "Transformations/OutputDataModule", "LHCbDIRAC.Core.Utilities.OutputDataPolicy"
        )

        threadTransformationClient = TransformationClient()
        threadTaskManager = LHCbWorkflowTasks(outputDataModule=outputDataModule, jobClass=LHCbJob)
        res.update({"TransformationClient": threadTransformationClient, "TaskManager": threadTaskManager})
        return res



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Test class for ModuleBase."""
# pylint: disable=protected-access, missing-docstring, invalid-name
import os
from itertools import product
from unittest.mock import MagicMock

import pytest

# mocks
from DIRAC.DataManagementSystem.Client.test.mock_DM import dm_mock
from LHCbDIRAC.BookkeepingSystem.Client.test.mock_BookkeepingClient import bkc_mock
from LHCbDIRAC.Workflow.Modules.mock_Commons import (
    version,
    prod_id,
    prod_job_id,
    wms_job_id,
    workflowStatus,
    stepStatus,
    step_id,
    step_number,
    step_commons,
    wf_commons,
)

# sut
from LHCbDIRAC.Workflow.Modules.ModuleBase import ModuleBase


def test__enableModule(mocker):
    mocker.patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())
    mb = ModuleBase(bkClientIn=bkc_mock, dm=dm_mock)
    mb.execute(
        prod_id,
        prod_job_id,
        wms_job_id,
        workflowStatus,
        stepStatus,
        wf_commons,
        step_commons[0],
        step_number,
        step_id,
    )
    assert mb._enableModule() is True


def test__checkLocalExistance(mocker):
    mocker.patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())
    mb = ModuleBase(bkClientIn=bkc_mock, dm=dm_mock)
    with pytest.raises(OSError):
        mb._checkLocalExistance(["aaa", "bbb"])

    with open("testFile.txt", "w"):
        res = mb._checkLocalExistance(["testFile.txt"])
        assert res == ["testFile.txt"]

        with open("testSecondFile.txt", "w"):
            res = mb._checkLocalExistance(["testFile.txt", "testSecondFile.txt"])
            assert sorted(res) == sorted(["testFile.txt", "testSecondFile.txt"])

            res = mb._checkLocalExistance(["TESTFILE.TXT", "testSecondFile.txt"])
            assert sorted(res) == sorted(["testFile.txt", "testSecondFile.txt"])
    os.remove("testFile.txt")
    os.remove("testSecondFile.txt")


candidateFiles = {
    "00012345_00012345_4.dst": {"lfn": "/lhcb/MC/2010/DST/123/123_45_4.dst", "type": "dst"},
    "00012345_00012345_2.digi": {"type": "digi"},
    "00012345_00012345_3.digi": {"type": "digi"},
    "00012345_00012345_5.AllStreams.dst": {
        "lfn": "/lhcb/MC/2010/DST/123/123_45_5.AllStreams.dst",
        "type": "allstreams.dst",
    },
    "00012345_00012345_1.sim": {"type": "sim"},
    "00038941_00000004_6.B2_D2.Strip.dst": {
        "lfn": "/lhcb/MC/2012/B2_D2.STRIP.DST/B2_D2.Strip.dst",
        "type": "B2_D2.strip.dst",
    },
    "Gauss_HIST_1.root": {"type": "GAUSSHIST"},
}

fileMasks = (
    ["dst"],
    "dst",
    ["sim"],
    ["digi"],
    ["digi", "sim"],
    "allstreams.dst",
    "B2_D2.strip.dst",
    [],
    ["B2_D2.strip.dst", "digi"],
    ["gausshist", "digi"],
)
stepMasks = (
    "",
    "5",
    "",
    ["2"],
    ["1", "3"],
    "",
    "",
    ["6"],
    [],
    ["1", "3"],
)

results = (
    {"00012345_00012345_4.dst": {"lfn": "/lhcb/MC/2010/DST/123/123_45_4.dst", "type": "dst"}},
    {},
    {"00012345_00012345_1.sim": {"type": "sim"}},
    {
        "00012345_00012345_2.digi": {"type": "digi"},
    },
    {"00012345_00012345_3.digi": {"type": "digi"}, "00012345_00012345_1.sim": {"type": "sim"}},
    {
        "00012345_00012345_5.AllStreams.dst": {
            "lfn": "/lhcb/MC/2010/DST/123/123_45_5.AllStreams.dst",
            "type": "allstreams.dst",
        }
    },
    {
        "00038941_00000004_6.B2_D2.Strip.dst": {
            "lfn": "/lhcb/MC/2012/B2_D2.STRIP.DST/B2_D2.Strip.dst",
            "type": "B2_D2.strip.dst",
        }
    },
    {
        "00038941_00000004_6.B2_D2.Strip.dst": {
            "lfn": "/lhcb/MC/2012/B2_D2.STRIP.DST/B2_D2.Strip.dst",
            "type": "B2_D2.strip.dst",
        }
    },
    {
        "00012345_00012345_2.digi": {"type": "digi"},
        "00012345_00012345_3.digi": {"type": "digi"},
        "00038941_00000004_6.B2_D2.Strip.dst": {
            "lfn": "/lhcb/MC/2012/B2_D2.STRIP.DST/B2_D2.Strip.dst",
            "type": "B2_D2.strip.dst",
        },
    },
    {"00012345_00012345_3.digi": {"type": "digi"}, "Gauss_HIST_1.root": {"type": "GAUSSHIST"}},
)

allCombinations = list(zip(fileMasks, stepMasks, results))


@pytest.mark.parametrize("fileMask, stepMask, result", allCombinations)
def test__applyMask(mocker, fileMask, stepMask, result):
    mocker.patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())
    mb = ModuleBase(bkClientIn=bkc_mock, dm=dm_mock)
    res = mb._applyMask(candidateFiles, fileMask, stepMask)
    assert res == result


def test__checkSanity(mocker):
    mocker.patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())
    mb = ModuleBase(bkClientIn=bkc_mock, dm=dm_mock)
    with pytest.raises(ValueError):
        mb._checkSanity(candidateFiles)


outputList = [
    {"outputDataType": "txt", "outputDataName": "foo_1.txt"},
    {"outputDataType": "py", "outputDataName": "bar_2.py"},
    {"outputDataType": "blah", "outputDataName": "myfoo.blah"},
]
outputLFNs = [
    "/lhcb/MC/2010/DST/00012345/0001/foo_1.txt",
    "/lhcb/MC/2010/DST/00012345/0001/bar_2.py",
    "/lhcb/user/f/fstagni/2015_06/prepend_myfoo.blah",
]

fileMasks = ["txt", ["txt", "py"], ["blah"], ["aa"], "", "", ""]
stepMasks = ["", None, None, None, "2", ["2", "3"], ["3"]]
results = [
    {"foo_1.txt": {"lfn": "/lhcb/MC/2010/DST/00012345/0001/foo_1.txt", "type": outputList[0]["outputDataType"]}},
    {
        "foo_1.txt": {"lfn": "/lhcb/MC/2010/DST/00012345/0001/foo_1.txt", "type": outputList[0]["outputDataType"]},
        "bar_2.py": {"lfn": "/lhcb/MC/2010/DST/00012345/0001/bar_2.py", "type": outputList[1]["outputDataType"]},
    },
    {"myfoo.blah": {"lfn": "/lhcb/user/f/fstagni/2015_06/prepend_myfoo.blah", "type": outputList[2]["outputDataType"]}},
    {},
    {"bar_2.py": {"lfn": "/lhcb/MC/2010/DST/00012345/0001/bar_2.py", "type": outputList[1]["outputDataType"]}},
    {"bar_2.py": {"lfn": "/lhcb/MC/2010/DST/00012345/0001/bar_2.py", "type": outputList[1]["outputDataType"]}},
    {},
]

allCombinations = list(zip(fileMasks, stepMasks, results))


@pytest.mark.parametrize("fileMask, stepMask, result", allCombinations)
def test_getCandidateFiles(mocker, fileMask, stepMask, result):
    mocker.patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())
    mb = ModuleBase(bkClientIn=bkc_mock, dm=dm_mock)

    with open("foo_1.txt", "w"):
        with open("bar_2.py", "w"):
            with open("myfoo.blah", "w"):
                res = mb.getCandidateFiles(outputList, outputLFNs, fileMask, stepMask)
                assert res == result

                with pytest.raises(ValueError):
                    mb.getCandidateFiles(
                        [
                            {"outputDataType": "txt", "outputDataName": "foo_1.txt"},
                            {"outputDataType": "py", "outputDataName": "bar_2.py"},
                            {"outputDataType": "blah", "outputDataName": "myfoo.blah"},
                        ],
                        [
                            "/lhcb/MC/2010/DST/00012345/0001/foo_1.txt",
                            "/lhcb/MC/2010/DST/00012345/0001/bar_2.py",
                            "/lhcb/user/f/fstagni/2015_06/prepend_myfoo_1.blah",
                        ],
                        "blah",
                        "",
                    )
                os.remove("foo_1.txt")
                os.remove("bar_2.py")
                os.remove("myfoo.blah")


inputDatas = ["previousStep", "previousStep", "LFN:123.raw"]
workflow_commons = [
    {
        "JobType": "User",
        "outputList": [
            {
                "stepName": "Brunel_1",
                "outputDataType": "brunelhist",
                "outputBKType": "BRUNELHIST",
                "outputDataName": "Brunel_00012345_00006789_1_Hist.root",
            },
            {
                "stepName": "Brunel_1",
                "outputDataType": "sdst",
                "outputBKType": "SDST",
                "outputDataName": "00012345_00006789_1.sdst",
            },
        ],
    },
    {
        "JobType": "User",
        "outputList": [
            {
                "stepName": "Brunel_1",
                "outputDataType": "brunelhist",
                "outputBKType": "BRUNELHIST",
                "outputDataName": "Brunel_00012345_00006789_1_Hist.root",
            },
            {"stepName": "Brunel_1", "outputDataType": "sdst", "outputBKType": "SDST", "outputDataName": "some.sdst"},
            {
                "stepName": "Brunel_1",
                "outputDataType": "sdst",
                "outputBKType": "SDST",
                "outputDataName": "00012345_00006789_1.sdst",
            },
        ],
    },
    {
        "JobType": "User",
        "outputList": [
            {
                "stepName": "Brunel_1",
                "outputDataType": "brunelhist",
                "outputBKType": "BRUNELHIST",
                "outputDataName": "Brunel_00012345_00006789_1_Hist.root",
            },
            {"stepName": "Brunel_1", "outputDataType": "sdst", "outputBKType": "SDST", "outputDataName": "some.sdst"},
            {
                "stepName": "Brunel_1",
                "outputDataType": "sdst",
                "outputBKType": "SDST",
                "outputDataName": "00012345_00006789_1.sdst",
            },
        ],
    },
]
results = [["00012345_00006789_1.sdst"], ["some.sdst", "00012345_00006789_1.sdst"], ["123.raw"]]

allCombinations = list(zip(inputDatas, workflow_commons, results))


@pytest.mark.parametrize("inputData, wf_c, result", allCombinations)
def test__determineStepInputData(mocker, inputData, wf_c, result):
    mocker.patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())
    mb = ModuleBase(bkClientIn=bkc_mock, dm=dm_mock)
    mb.stepName = "DaVinci_2"

    mb.gaudiSteps = ["Brunel_1", "DaVinci_2"]
    mb.workflow_commons = wf_c
    mb.inputDataType = "SDST"
    res = mb._determineStepInputData(inputData)
    assert res == result


@pytest.mark.parametrize("s_cs", list(step_commons))
def test__determineOutputs(mocker, s_cs):
    mocker.patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())
    mb = ModuleBase(bkClientIn=bkc_mock, dm=dm_mock)
    mb.jobType = "merge"
    mb.step_id = "00000123_00000456_1"
    mb.stepInputData = ["foo", "bar"]
    mb.step_commons = dict(s_cs)
    mb.step_commons["listoutput"] = [
        {"outputDataType": "bhadron.dst;sdst", "outputDataName": "00000123_00000456_1.bhadron.dst;sdst"}
    ]
    outF, outft, histos = mb._determineOutputs()
    assert outF == [{"outputDataType": "sdst", "outputDataName": "00000123_00000456_1.sdst", "outputBKType": "SDST"}]
    assert outft == ["sdst"]
    assert histos is False

    mb.step_commons["listoutput"] = [
        {"outputDataType": "root", "outputDataName": "00000123_00000456_1.root", "outputBKType": "ROOT"}
    ]
    outF, outft, histos = mb._determineOutputs()
    assert outF == [{"outputDataType": "root", "outputDataName": "00000123_00000456_1.root", "outputBKType": "ROOT"}]
    assert outft == ["root"]
    assert histos is False

    mb.jobType = "reco"
    mb.step_commons = s_cs
    mb.step_commons["listoutput"] = [
        {"outputDataType": "sdst", "outputDataName": "00000123_00000456_1.sdst", "outputBKType": "SDST"}
    ]
    outF, outft, histos = mb._determineOutputs()
    assert outF == [{"outputDataType": "sdst", "outputDataName": "00000123_00000456_1.sdst", "outputBKType": "SDST"}]
    assert outft == ["sdst"]
    assert histos is False


def test__findOutputs(mocker):
    mocker.patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())
    mb = ModuleBase(bkClientIn=bkc_mock, dm=dm_mock)

    with pytest.raises(IOError):
        with open("aaa.Bhadron.dst", "w"):
            with open("ccc.charm.mdst", "w"):
                with open("prova.txt", "w"):
                    stepOutput = [
                        {"outputDataType": "BHADRON.DST", "outputDataName": "aaa.bhadron.dst"},
                        {"outputDataType": "CALIBRATION.DST", "outputDataName": "bbb.calibration.dst"},
                        {"outputDataType": "CHARM.MDST", "outputDataName": "ccc.charm.mdst"},
                        {
                            "outputDataType": "CHARMCONTROL.DST",
                            "outputDataName": "00012345_00012345_2.CHARMCONTROL.DST",
                        },
                        {"outputDataType": "CHARMFULL.DST", "outputDataName": "00012345_00012345_2.CHARMFULL.DST"},
                        {"outputDataType": "LEPTONIC.MDST", "outputDataName": "00012345_00012345_2.LEPTONIC.MDST"},
                        {
                            "outputDataType": "LEPTONICFULL.DST",
                            "outputDataName": "00012345_00012345_2.LEPTONICFULL.DST",
                        },
                        {"outputDataType": "MINIBIAS.DST", "outputDataName": "00012345_00012345_2.MINIBIAS.DST"},
                        {"outputDataType": "RADIATIVE.DST", "outputDataName": "00012345_00012345_2.RADIATIVE.DST"},
                        {
                            "outputDataType": "SEMILEPTONIC.DST",
                            "outputDataName": "00012345_00012345_2.SEMILEPTONIC.DST",
                        },
                        {"outputDataType": "HIST", "outputDataName": "DaVinci_00012345_00012345_2_Hist.root"},
                    ]
                    mb._findOutputs(stepOutput)
                    os.remove("aaa.Bhadron.dst")
                    os.remove("ccc.charm.mdst")
                    os.remove("prova.txt")

    stepOutput = [{"outputDataType": "BHADRON.DST", "outputDataName": "aaa.bhadron.dst"}]
    outExp = [
        {
            "outputDataType": "bhadron.dst",
            "outputBKType": "BHADRON.DST",
            "outputDataName": "aaa.Bhadron.dst",
            "stepName": "someApp_1",
        }
    ]
    bkExp = ["BHADRON.DST"]
    mb.stepName = "someApp_1"
    out, bk = mb._findOutputs(stepOutput)
    assert out == outExp
    assert bk == bkExp


def test_getFileMetadata(mocker):
    mocker.patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())
    mb = ModuleBase(bkClientIn=bkc_mock, dm=dm_mock)

    candidateFiles = {
        "foo_1.txt": {"lfn": "/lhcb/MC/2010/DST/00012345/0001/foo_1.txt", "type": "txt", "workflowSE": "SE1"},
        "bar_2.py": {"lfn": "/lhcb/MC/2010/DST/00012345/0001/bar_2.py", "type": "py", "workflowSE": "SE2"},
    }

    expectedResult = {
        "bar_2.py": {
            "filedict": {
                "Status": "Waiting",
                "LFN": "/lhcb/MC/2010/DST/00012345/0001/bar_2.py",
                "GUID": "D41D8CD9-8F00-B204-E980-0998ECF8427E",
                "Checksum": "000001",
                "ChecksumType": "ADLER32",
                "Size": 0,
            },
            "lfn": "/lhcb/MC/2010/DST/00012345/0001/bar_2.py",
            "workflowSE": "SE2",
            "localpath": os.getcwd() + "/bar_2.py",
            "guid": "D41D8CD9-8F00-B204-E980-0998ECF8427E",
            "type": "py",
        },
        "foo_1.txt": {
            "filedict": {
                "Status": "Waiting",
                "LFN": "/lhcb/MC/2010/DST/00012345/0001/foo_1.txt",
                "GUID": "D41D8CD9-8F00-B204-E980-0998ECF8427E",
                "Checksum": "000001",
                "ChecksumType": "ADLER32",
                "Size": 0,
            },
            "lfn": "/lhcb/MC/2010/DST/00012345/0001/foo_1.txt",
            "workflowSE": "SE1",
            "localpath": os.getcwd() + "/foo_1.txt",
            "guid": "D41D8CD9-8F00-B204-E980-0998ECF8427E",
            "type": "txt",
        },
    }

    with open("foo_1.txt", "w"):
        with open("bar_2.py", "w"):
            result = mb.getFileMetadata(candidateFiles)
            assert result == expectedResult
            os.remove("foo_1.txt")
            os.remove("bar_2.py")



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Fix incorrect file GUIDs."""
import os

import DIRAC
from DIRAC.Core.Base.Script import Script

keep = None
localFile = None


def leave(msg, error=None, exitCode=0):
    from DIRAC import gLogger

    if error is None:
        error = {}

    if not keep and os.path.exists(localFile):
        os.remove(localFile)
        gLogger.info(f"Local file {localFile} removed")

    if error:
        errMsg = error.get(
            "Message",
            list(error.get("Value", {}).get("Failed", {"": "Unknown reason"}).values())[0],
        )
        gLogger.error(msg, f": {errMsg}")
    else:
        gLogger.always(msg)
    DIRAC.exit(exitCode)


@Script()
def main():
    global keep
    global localFile

    Script.registerSwitch("f:", "OldLFN=", "LFN of existing file to be fixed.")
    Script.registerSwitch(
        "n:", "NewLFN=", "Optional: specify a new LFN for the file (retaining the existing file with incorrect GUID)."
    )
    Script.registerSwitch("D:", "Directory=", "Optional: directory to download file (defaults to TMPDIR then PWD).")
    Script.registerSwitch("k", "Keep", "Optional: specify this switch to retain the local copy of the downloaded file")
    Script.registerSwitch(
        "m", "SafeMode", "Optional: specify this switch to run the script in safe mode (will check the GUIDs only)"
    )
    Script.setUsageMessage("\n".join([__doc__, "Usage:", f"  {Script.scriptName} [option|cfgfile] [OldLFN]"]))
    Script.parseCommandLine(ignoreErrors=True)

    args = Script.getPositionalArgs()

    from DIRAC.Interfaces.API.Dirac import Dirac
    from LHCbDIRAC.Core.Utilities.File import makeGuid

    from DIRAC import gLogger

    oldLFN = ""
    newLFN = ""
    localFile = ""
    keep = False
    safe = False
    directory = os.getcwd()
    if "TMPDIR" in os.environ:
        directory = os.environ["TMPDIR"]

    dirac = Dirac()

    if args:
        oldLFN = args.pop(0)

    for switch in Script.getUnprocessedSwitches():
        if switch[0].lower() in ("f", "oldlfn"):
            oldLFN = switch[1]
        elif switch[0].lower() in ("n", "newlfn"):
            newLFN = switch[1]
        elif switch[0].lower() in ("d", "directory"):
            directory = switch[1]
        elif switch[0].lower() in ("k", "keep"):
            keep = True
        elif switch[0].lower() in ("m", "safemode"):
            safe = True

    oldLFN = oldLFN.replace("LFN:", "")
    newLFN = newLFN.replace("LFN:", "")

    if not oldLFN:
        leave("The original LFN of the file to be fixed must be specified", exitCode=2)

    if not os.path.exists(directory):
        leave(f"Optional directory {directory} doesn't exist", exitCode=2)

    if not newLFN:
        gLogger.verbose(f"No new LFN specified, will replace the existing LFN {oldLFN}")
        newLFN = oldLFN

    gLogger.verbose(f"Directory for downloading file is set to {directory}")

    replicaInfo = dirac.getReplicas(oldLFN)
    if not replicaInfo["OK"] or replicaInfo["Value"]["Failed"]:
        leave(f"Could not get replica information for {oldLFN}", replicaInfo, exitCode=2)

    replicas = replicaInfo["Value"]["Successful"][oldLFN]
    storageElements = list(replicas)
    if not storageElements:
        leave(f"Could not determine SEs for replicas of {oldLFN}", exitCode=2)

    gLogger.info(f"Existing LFN has replicas at: {', '.join(storageElements)}")

    oldGUID = dirac.getLfnMetadata(oldLFN)
    if not oldGUID["OK"] or oldGUID["Value"]["Failed"]:
        leave("Could not obtain GUID from FC for %s - %s" % oldLFN, oldGUID, exitCode=2)
    oldGUID = oldGUID["Value"]["Successful"][oldLFN]["GUID"]
    gLogger.verbose(f"Existing GUID is {oldGUID}")

    # retrieve original file
    localFile = os.path.join(directory, os.path.basename(oldLFN))
    if not os.path.exists(localFile):
        download = dirac.getFile(oldLFN, directory)
        if not download["OK"] or download["Value"]["Failed"]:
            leave(f"Could not download file with message - {download['Message']}", download, exitCode=2)
    else:
        gLogger.always(f"Found file {os.path.basename(oldLFN)} in local directory, will not re-download")

    newGUID = makeGuid(localFile)[localFile]

    if newGUID == oldGUID:
        leave(f"Old and new GUIDs have the same value ({oldGUID}), exiting without changes")

    if safe:
        leave(f"Safe mode - found file GUID = {newGUID} and existing GUID = {oldGUID}, exiting without changes")

    gLogger.verbose(f"Will set old GUID to {newGUID} from {oldGUID}")
    if newLFN == oldLFN:
        gLogger.always("Removing old LFN from storages before adding new LFN")
        result = dirac.removeFile(oldLFN)
        if not result["OK"]:
            leave("Could not remove existing LFN from Grid storage", result, exitCode=2)

    gLogger.always(f"Uploading {localFile} as LFN:{newLFN} with replica at {storageElements[0]} and GUID = {newGUID}")
    result = dirac.addFile(newLFN, localFile, storageElements[0], fileGuid=newGUID, printOutput=False)
    if not result["OK"]:
        leave(f"Failed to copy and register new LFN:{newLFN}", result, exitCode=2)

    leave("")


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Test the logic to detect successful and failed file accesses via xroot.

We use test pool and summary xml files that correspond to 3 LFNs:
* one that worked at the first attempt
* one that worked at the second attempt
* one that did not work at all
"""
import os
from collections import defaultdict
from unittest.mock import MagicMock, patch

from LHCbDIRAC.Core.Utilities.XMLSummaries import XMLSummary
from DIRAC.Resources.Catalog.PoolXMLCatalog import PoolXMLCatalog

from LHCbDIRAC.Workflow.Modules.AnalyseFileAccess import AnalyseFileAccess

testdir = os.path.dirname(__file__)
poolFile = os.path.join(testdir, "pool_xml_catalog.xml")
summaryFile = os.path.join(testdir, "summary.xml")


@patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())
def test_analyseFileAccess(mockRequestValidator):
    """Analyze the file accesses from a pool xml catalog and the xml summary."""

    xmlCatalog = PoolXMLCatalog(xmlfile=poolFile)
    xmlSummary = XMLSummary(summaryFile)

    fileAccessAnalyzer = AnalyseFileAccess()
    accessAttempts = fileAccessAnalyzer._checkFileAccess(xmlCatalog, xmlSummary)

    # Count how many successes
    accessPerSE = defaultdict(lambda: defaultdict(int))

    for se, success in accessAttempts:
        accessPerSE[se][success] += 1

    # 'NEVERUSED-DST' should not have any counter
    assert "NEVERUSED-DST" not in accessPerSE

    # GOOD-DST should have 2 good reads and no bad
    assert False not in accessPerSE["GOOD-DST"]
    assert accessPerSE["GOOD-DST"][True] == 2

    # BAD-DST should have 2 bad reads and no good
    assert True not in accessPerSE["BAD-DST"]
    assert accessPerSE["BAD-DST"][False] == 2

    # OTHERBAD-DST should have no good read and 1 bad read
    assert True not in accessPerSE["OTHERBAD-DST"]
    assert accessPerSE["OTHERBAD-DST"][False] == 1



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

from unittest.mock import Mock
import unittest

from LHCbDIRAC.TransformationSystem.Utilities.StateMachine import TransformationFilesStateMachine


class UtilitiesTestCase(unittest.TestCase):
    def setUp(self):
        self.transClient = Mock()

    def tearDown(self):
        pass


class tsfmSuccess(UtilitiesTestCase):
    def test_setState(self):
        tsfm = TransformationFilesStateMachine("Unused")
        res = tsfm.setState("Assigned")
        self.assertTrue(res["OK"])
        self.assertEqual(res["Value"], "Assigned")

        tsfm = TransformationFilesStateMachine("Unused")
        res = tsfm.setState("Unused")
        self.assertTrue(res["OK"])
        self.assertEqual(res["Value"], "Unused")

        tsfm = TransformationFilesStateMachine("Unused")
        res = tsfm.setState("MissingInFC")
        self.assertTrue(res["OK"])
        self.assertEqual(res["Value"], "MissingInFC")

        tsfm = TransformationFilesStateMachine("Unused")
        res = tsfm.setState("Processed")
        self.assertTrue(res["OK"])
        self.assertEqual(res["Value"], "Processed")

        tsfm = TransformationFilesStateMachine("Unused")
        res = tsfm.setState("Processed")
        self.assertTrue(res["OK"])
        self.assertEqual(res["Value"], "Processed")
        res = tsfm.setState("Processed")
        self.assertTrue(res["OK"])
        self.assertEqual(res["Value"], "Processed")


class tsfmFailure(UtilitiesTestCase):
    def test_setState(self):
        tsfm = TransformationFilesStateMachine("Unused")
        res = tsfm.setState("Assigne")
        self.assertFalse(res["OK"])

        tsfm = TransformationFilesStateMachine("Processed")
        res = tsfm.setState("Assigne")
        self.assertFalse(res["OK"])

        tsfm = TransformationFilesStateMachine("Processed")
        res = tsfm.setState("Processed")
        self.assertTrue(res["OK"])

        tsfm = TransformationFilesStateMachine("Processed")
        res = tsfm.setState("Unused")
        self.assertTrue(res["OK"])
        self.assertEqual(res["Value"], "Processed")

        tsfm = TransformationFilesStateMachine("MaxReset")
        res = tsfm.setState("Processed")
        self.assertTrue(res["OK"])
        self.assertEqual(res["Value"], "MaxReset")


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(UtilitiesTestCase)
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(tsfmSuccess))
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Unit tests for Workflow Module Uploads."""
# pylint: disable=protected-access, missing-docstring
import os
import copy
import json

from unittest.mock import MagicMock
import pytest

from DIRAC import gConfig, gLogger
from DIRAC.RequestManagementSystem.Client.Request import Request
from DIRAC.RequestManagementSystem.Client.Operation import Operation
from DIRAC.RequestManagementSystem.Client.File import File

# mocks
from DIRAC.DataManagementSystem.Client.test.mock_DM import dm_mock
from DIRAC.Resources.Catalog.test.mock_FC import fc_mock

from LHCbDIRAC.Workflow.Modules.mock_Commons import (
    prod_id,
    prod_job_id,
    wms_job_id,
    workflowStatus,
    stepStatus,
    step_id,
    step_number,
    step_commons,
    wf_commons,
    rc_mock,
)
from LHCbDIRAC.BookkeepingSystem.Client.test.mock_BookkeepingClient import bkc_mock

# sut
from LHCbDIRAC.Workflow.Modules.UploadOutputData import UploadOutputData


ft_mock = MagicMock()
ft_mock.transferAndRegisterFile.return_value = {"OK": True, "Value": {"uploadedSE": ""}}
ft_mock.transferAndRegisterFileFailover.return_value = {"OK": True, "Value": {}}
ft_mock.request = rc_mock
ft_mock.FileCatalog = fc_mock


@pytest.fixture
def rmFiles():
    yield
    for fileProd in [
        "appLog",
        "foo.txt",
        "bar.txt",
        "aaa.Bhadron.dst",
        "bbb.Calibration.dst",
        "bar.py",
        "aLongLog.log",
        "bookkeeping_123_00000456_321.xml",
        "aLongLog.log.gz",
        "ccc.charm.mdst",
        "ccc.charm.mdst",
        "prova.txt",
        "aLog.log",
        "BAR.txt",
        "FooBAR.ext.txt",
        "foo_1.txt",
        "bar_2.py",
        "ErrorLogging_Step1_coredump.log",
        "123_00000456_request.xml",
        "lfn1",
        "lfn2",
        "XMLSummaryFile",
        "aaa.bhadron.dst",
        "bbb.calibration.dst",
        "ProductionOutputData",
        "data.py",
        "123_00000456_request.json",
        "00000123_00000456.tar",
        "someOtherDir",
        "DISABLE_WATCHDOG_CPU_WALLCLOCK_CHECK",
        "myfoo.blah",
        "prodConf_someApp__.py",
        "prodConf_someApp___.py",
    ]:
        try:
            os.remove(fileProd)
        except OSError:
            continue


def test_execute(mocker, rmFiles, monkeypatch):
    mocker.patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())
    mocker.patch("LHCbDIRAC.Workflow.Modules.UploadOutputData.FileCatalog", side_effect=fc_mock)

    monkeypatch.setattr(gConfig, "getOption", lambda x, y: {"OK": True, "Value": "exp1"})

    uod = UploadOutputData(bkClient=bkc_mock, dm=dm_mock)
    uod.log = gLogger
    uod.log.setLevel("DEBUG")
    uod.siteName = "DIRAC.Test.ch"
    uod.failoverTransfer = ft_mock

    # no errors, no input data
    for wf_cs in copy.deepcopy(wf_commons[0:1]):
        if "outputList" in wf_cs:
            continue
        for s_cs in step_commons[0:1]:
            fileDescendants = {}
            assert (
                uod.execute(
                    prod_id,
                    prod_job_id,
                    wms_job_id,
                    workflowStatus,
                    stepStatus,
                    wf_cs,
                    s_cs,
                    step_number,
                    step_id,
                    SEs=["SomeSE"],
                    fileDescendants=fileDescendants,
                )["OK"]
                is True
            )

    # no errors, input data
    for wf_cs in copy.deepcopy(wf_commons):
        for s_cs in step_commons:
            for transferAndRegisterFile in (
                {"OK": True, "Value": {"uploadedSE": ""}},
                {"OK": False, "Message": "error"},
            ):
                # for transferAndRegisterFileFailover in ( {'OK': True, 'Value': {}},
                # {'OK': False, 'Message': 'error'} ):
                ft_mock.transferAndRegisterFile.return_value = transferAndRegisterFile
                open("foo.txt", "w").close()
                open("bar.txt", "w").close()
                if "InputData" not in wf_cs:
                    continue
                if wf_cs["InputData"] == "":
                    continue
                wf_cs["outputList"] = [
                    {"outputDataType": "txt", "outputDataName": "foo.txt"},
                    {"outputDataType": "txt", "outputDataName": "bar.txt"},
                ]
                wf_cs["ProductionOutputData"] = [
                    "/lhcb/MC/2010/DST/00012345/0001/foo.txt",
                    "/lhcb/MC/2010/DST/00012345/0001/bar.txt",
                ]
                # bkc_mock.getFileDescendents.return_value = {
                #     'OK': False,
                #     'rpcStub': (
                #         ( 'Bookkeeping/BookkeepingManager',
                #         {'skipCACheck': False,
                #         'timeout': 3600} ),
                #         'getFileDescendents', ( ['foo'], 9, 0, True ) ),
                #     'Value': {'Successful': {'foo.txt': ['baaar']}, 'Failed': [], 'NotProcessed': []}
                # }
                fileDescendants = {"foo.txt": ["baaar"]}
                assert (
                    uod.execute(
                        prod_id,
                        prod_job_id,
                        wms_job_id,
                        workflowStatus,
                        stepStatus,
                        wf_cs,
                        s_cs,
                        step_number,
                        step_id,
                        SEs=["SomeSE"],
                        fileDescendants=fileDescendants,
                    )["OK"]
                    is False
                )

                # bkc_mock.getFileDescendents.return_value = {
                #     'OK': True,
                #     'rpcStub': ( (
                #         'Bookkeeping/BookkeepingManager',
                #         {'skipCACheck': False, 'timeout': 3600} ),
                #         'getFileDescendents', ( ['foo'], 9, 0, True ) ),
                #     'Value': {'Successful': {}, 'Failed': [], 'NotProcessed': []},
                # }
                if wf_cs["Request"] == "":
                    continue
                fileDescendants = {}
                res = uod.execute(
                    prod_id,
                    prod_job_id,
                    wms_job_id,
                    workflowStatus,
                    stepStatus,
                    wf_cs,
                    s_cs,
                    step_number,
                    step_id,
                    SEs=["SomeSE"],
                    fileDescendants=fileDescendants,
                )
                assert res["OK"] is True
                #            if transferAndRegisterFileFailover['OK']:
                #              self.assertTrue( res['OK'] )
                #            else:
                #              self.assertFalse( res['OK'] )
                os.remove("foo.txt")
                os.remove("bar.txt")


def test__getLFNsForBKRegistration(mocker):
    mocker.patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())

    f1 = File()
    f1.LFN = "/a/1.txt"
    f2 = File()
    f2.LFN = "/a/2.txt"
    f3 = File()
    f3.LFN = "/a/3.txt"

    o1 = Operation()
    o1.Type = "RegisterFile"
    o1.addFile(f1)
    o2 = Operation()
    o2.Type = "RegisterFile"
    o2.addFile(f2)
    o3 = Operation()
    o3.Type = "ForwardDISET"
    o4 = Operation()
    o4.Type = "RegisterFile"
    o4.addFile(f1)
    o4.addFile(f3)

    r = Request()
    r.addOperation(o4)
    r.addOperation(o1)
    r.addOperation(o2)
    r.addOperation(o3)

    uod = UploadOutputData(bkClient=bkc_mock, dm=dm_mock)
    uod.request = r

    lfns = {"/a/1.txt", "/a/5.txt", "/a/6.txt", "/a/3.txt"}
    lfnsForBkReg = uod._getLFNsForBKRegistration(lfns)
    assert sorted(lfnsForBkReg) == sorted(["/a/5.txt", "/a/6.txt"])


def test__cleanUp(mocker):
    mocker.patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())

    f1 = File()
    f1.LFN = "/a/1.txt"
    f2 = File()
    f2.LFN = "/a/2.txt"
    f3 = File()
    f3.LFN = "/a/3.txt"

    o1 = Operation()
    o1.Type = "RegisterFile"
    o1.addFile(f1)
    o2 = Operation()
    o2.Type = "RegisterFile"
    o2.addFile(f2)
    o3 = Operation()
    o3.Type = "ForwardDISET"
    o4 = Operation()
    o4.Type = "RegisterFile"
    o4.addFile(f1)
    o4.addFile(f3)

    r = Request()
    r.addOperation(o4)
    r.addOperation(o1)
    r.addOperation(o2)
    r.addOperation(o3)

    uod = UploadOutputData(bkClient=bkc_mock, dm=dm_mock)
    uod.failoverTransfer = ft_mock
    uod.request = r

    expected = Request()
    expected.addOperation(o3)
    removeOp = Operation()
    removeOp.Type = "RemoveFile"
    fileRemove1 = File()
    fileRemove1.LFN = "/a/1.txt"
    fileRemove2 = File()
    fileRemove2.LFN = "/a/2.txt"
    fileRemove3 = File()
    fileRemove3.LFN = "/a/notPresent.txt"
    removeOp.addFile(fileRemove1)
    removeOp.addFile(fileRemove3)
    removeOp.addFile(fileRemove2)
    expected.addOperation(removeOp)

    uod._cleanUp(
        {"1.txt": {"lfn": "/a/1.txt"}, "2.txt": {"lfn": "/a/2.txt"}, "notPresent.txt": {"lfn": "/a/notPresent.txt"}}
    )

    for opsR, opsE in zip(uod.request, expected):
        opsRLoaded = json.loads(str(opsR))
        opsELoaded = json.loads(str(opsE))
        for k in set(opsRLoaded) | set(opsELoaded):
            if k == "Files":
                assert sorted(opsRLoaded[k], key=lambda x: x["LFN"]) == sorted(opsELoaded[k], key=lambda x: x["LFN"])
            else:
                assert opsRLoaded[k] == opsELoaded[k]



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Module that contains client access to the transformation DB handler.

This is a very simple extension to the DIRAC one
"""
from DIRAC import S_OK, gLogger
from DIRAC.Core.Utilities.JEncode import strToIntDict
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient as DIRACTransformationClient
from LHCbDIRAC.ProductionManagementSystem.Utilities.StateMachine import ProductionsStateMachine
from LHCbDIRAC.TransformationSystem.Utilities.StateMachine import TransformationFilesStateMachine


class TransformationClient(DIRACTransformationClient):
    """Exposes the functionality available in the
    LHCbDIRAC/TransformationHandler.

    This inherits the DIRAC base Client for direct execution of server functionality.
    The following methods are available (although not visible here).

    BK query manipulation
        deleteBookkeepingQuery(queryID)
        deleteTransformationBookkeepingQuery(transName)
        addBookkeepingQuery(transID,queryDict)
        getBookkeepingQuery(transName)
    """

    def __init__(self, **kwargs):
        DIRACTransformationClient.__init__(self, **kwargs)
        self.dataProcessingTypes = Operations().getValue("Transformations/DataProcessing", [])

    def addTransformation(
        self,
        transName,
        description,
        longDescription,
        transfType,
        plugin,
        agentType,
        fileMask,
        transformationGroup="General",
        groupSize=1,
        inheritedFrom=0,
        body="",
        maxTasks=0,
        eventsPerTask=0,
        addFiles=True,
        inputMetaQuery=None,
        outputMetaQuery=None,
        bkQuery=None,
        timeout=1800,
    ):
        res = super().addTransformation(
            transName,
            description,
            longDescription,
            transfType,
            plugin,
            agentType,
            fileMask,
            transformationGroup=transformationGroup,
            groupSize=groupSize,
            inheritedFrom=inheritedFrom,
            body=body,
            maxTasks=maxTasks,
            eventsPerTask=eventsPerTask,
            addFiles=addFiles,
            inputMetaQuery=inputMetaQuery,
            outputMetaQuery=outputMetaQuery,
            timeout=timeout,
        )
        if not res["OK"]:
            return res
        transID = res["Value"]
        if bkQuery:
            res = self._getRPC().addBookkeepingQuery(transID, bkQuery)
            if not res["OK"]:
                gLogger.error("Failed to publish BKQuery for transformation", f"{transID} {res['Message']}")
                return res
        return S_OK(transID)

    def getDestinationForRun(self, runIDs):
        retVal = self._getRPC().getDestinationForRun(runIDs)
        if not retVal["OK"]:
            return retVal
        return S_OK(strToIntDict(retVal["Value"]))

    def _applyTransformationStatusStateMachine(self, transIDAsDict, dictOfProposedstatus, force):
        """Performs a state machine check for productions when asked to change the
        status."""
        originalStatus, transformationType = list(transIDAsDict.values())[0][0:2]
        proposedStatus = list(dictOfProposedstatus.values())[0]
        if force:
            return proposedStatus
        else:
            if transformationType in self.dataProcessingTypes:
                stateChange = ProductionsStateMachine(originalStatus).setState(proposedStatus)
                if not stateChange["OK"]:
                    return originalStatus
                else:
                    return stateChange["Value"]
            else:
                return proposedStatus

    def _applyTransformationFilesStateMachine(self, tsFilesAsDict, dictOfProposedLFNsStatus, force):
        """Apply LHCb state machine for transformation files."""
        newStatuses = dict()
        for lfn, newStatus in dictOfProposedLFNsStatus.items():
            if lfn in tsFilesAsDict:
                currentStatus = tsFilesAsDict[lfn][0]
                if force:
                    # We do whatever is requested
                    newStatus = dictOfProposedLFNsStatus[lfn]
                else:
                    # Special case for Assigned -> Unused
                    if currentStatus.lower() == "assigned" and newStatus.lower() == "unused":
                        errorCount = tsFilesAsDict[lfn][1]
                        if errorCount and ((errorCount % self.maxResetCounter) == 0):
                            newStatus = "MaxReset"

                    tfsm = TransformationFilesStateMachine(currentStatus)
                    stateChange = tfsm.setState(newStatus)
                    if stateChange["OK"]:
                        newStatus = stateChange["Value"]
                # Only worth setting the status if different from current one
                if newStatus.lower() != currentStatus.lower():
                    newStatuses[lfn] = newStatus

        return newStatuses



###############################################################################
# (c) Copyright 2022 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
import os
import socket

import psutil
import xml.etree.ElementTree as ET
from pydantic import BaseModel as _BaseModel

import LHCbDIRAC


class BaseModel(_BaseModel, extra="forbid"):
    pass


DEFAULT_WNCPUPOWER = "unknown"
DEFAULT_WorkerNode = "unknown"
DEFAULT_WNMEMORY = "unknown"
DEFAULT_WNMODEL = "unknown"
DEFAULT_WNCACHE = "unknown"
try:
    DEFAULT_WNCPUPOWER = str(psutil.cpu_freq()[0])
    DEFAULT_WorkerNode = os.environ.get("HOSTNAME", os.environ.get("HOST", socket.gethostname()))
    DEFAULT_WNMEMORY = str(psutil.virtual_memory()[1] // 1024)
    with open("/proc/cpuinfo") as cpuinfo:
        info = cpuinfo.readlines()
    DEFAULT_WNMODEL = [x.strip().split(":")[1] for x in info if "model name" in x][0].strip()
    DEFAULT_WNCACHE = [x.strip().split(":")[1] for x in info if "cache size" in x][0].strip()
except Exception as x:
    pass


class BookkeepingJobInfo(BaseModel):
    ConfigName: str
    ConfigVersion: str
    Date: str
    Time: str

    class TypedParameters(BaseModel):
        CPUTIME: str | None = None
        ExecTime: str | None = None
        WNMODEL: str = DEFAULT_WNMODEL
        WNCPUPOWER: str = DEFAULT_WNCPUPOWER
        WNCACHE: str = DEFAULT_WNCACHE
        WorkerNode: str = DEFAULT_WorkerNode
        WNMEMORY: str = DEFAULT_WNMEMORY
        WNCPUHS06: str | None = None
        WNMJFHS06: str | None = None
        NumberOfProcessors: str | None = None
        Production: str | None = None
        DiracJobId: str | None = None
        Name: str | None = None
        JobStart: str | None = None
        JobEnd: str | None = None
        Location: str | None = None
        JobType: str | None = None
        ProgramName: str | None = None
        ProgramVersion: str | None = None
        DiracVersion: str = LHCbDIRAC.__version__
        FirstEventNumber: str | None = None
        StatisticsRequested: str | None = None
        StepID: str | None = None
        NumberOfEvents: str | None = None

    typed_parameters: TypedParameters = TypedParameters()

    input_files: list[str] = []

    class OutputFile(BaseModel):
        Name: str
        TypeName: str
        TypeVersion: str
        EventTypeId: str | None = None
        EventStat: str | None = None
        FileSize: str
        CreationDate: str
        MD5Sum: str
        Guid: str
        Replica: str | None = None

    output_files: list[OutputFile] = []

    simulation_condition: str | None = None

    def to_xml(self):
        root = ET.Element(
            "Job",
            ConfigName=self.ConfigName,
            ConfigVersion=self.ConfigVersion,
            Date=self.Date,
            Time=self.Time,
        )
        root.extend(
            [
                ET.Element("TypedParameter", Name=k, Value=str(v), Type="Info")
                for k, v in self.typed_parameters.model_dump().items()
                if v is not None
            ]
        )
        root.extend([ET.Element("InputFile", Name=lfn) for lfn in self.input_files])
        for output_file in self.output_files:
            data = dict(output_file.dict())
            output_file = ET.Element(
                "OutputFile",
                Name=data.pop("Name"),
                TypeName=data.pop("TypeName"),
                TypeVersion=data.pop("TypeVersion"),
            )
            if replica := data.pop("Replica"):
                output_file.append(ET.Element("Replica", Name=replica, Location="Web"))
            output_file.extend(
                [ET.Element("Parameter", Name=k, Value=str(v)) for k, v in data.items() if v is not None]
            )
            root.append(output_file)
        if self.simulation_condition is not None:
            simulation_condition = ET.Element("SimulationCondition")
            simulation_condition.append(ET.Element("Parameter", Name="SimDescription", Value=self.simulation_condition))
        ET.indent(root, "    ")
        return b"\n".join(
            [
                b'<?xml version="1.0" encoding="ISO-8859-1"?>\n',
                b'<!DOCTYPE Job SYSTEM "book.dtd">\n',
                ET.tostring(root, encoding="ISO-8859-1", method="xml", xml_declaration=False),
            ]
        )



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
Split a bookkeeping path into run ranges containing more or less a certain number of files.
This is used to prepare the (in)famous production spreadsheet

We expect that a one and only one event type be provided
"""

from collections import defaultdict
from DIRAC.Core.Base.Script import Script


def take_run_while(run_files, chunk_size):
    """
    Return list of run number as long as the total number of files
    is not above a certain threshold
    Of course, it can overshoot...

    :param run_files: run_number: nb of files in it
    :param chunk_size: max sum

    :return: run list, sum

    """

    fileSum = 0
    first_idx = 0

    runs = sorted(run_files)
    for i, run in enumerate(runs):
        fileSum += run_files[run]
        if fileSum > chunk_size:
            chunk = runs[first_idx : i + 1]
            first_idx = i + 1
            chunkSum = fileSum
            fileSum = 0
            yield chunk, chunkSum
    yield runs[first_idx:], sum([run_files[run] for run in runs[first_idx:]])


def executeSplitRunsForProd(bkQuery, chunk_size):
    from DIRAC import gLogger, exit as diracExit

    from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

    bkClient = BookkeepingClient()

    if not bkQuery:
        gLogger.notice("No BK query given...")
        diracExit(1)

    runList = sorted(bkQuery.getBKRuns())
    requested_event_type = bkQuery.getBKEventTypes()

    if len(requested_event_type) != 1:
        gLogger.notice("Only one event type supported")
        diracExit(1)

    requested_event_type = int(requested_event_type[0])

    requested_cond = bkQuery.getConditions()

    cond_runs = defaultdict(dict)
    for run in runList:
        res = bkClient.getRunInformations(run)
        if not res["OK"]:
            gLogger.notice(f"Error getting info for run {run}:{res['Message']}")
            continue
        runInfo = res["Value"]
        actual_cond = runInfo["DataTakingDescription"]
        if requested_cond and actual_cond != requested_cond:
            continue
        try:
            nb_of_files = runInfo["Number of file"][runInfo["Stream"].index(requested_event_type)]
            cond_runs[actual_cond][run] = nb_of_files
            pass
        # No files of the requested event type
        except ValueError:
            continue

    for cond, runsDict in cond_runs.items():
        for run_list, nb_of_files in take_run_while(runsDict, chunk_size):
            print(f"{cond}:{run_list[0]}:{run_list[-1]}:{nb_of_files}")


@Script()
def main():
    from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript

    dmScript = DMScript()
    dmScript.registerBKSwitches()
    Script.registerSwitch("", "ChunkSize=", "   Size of the chunks (default 50000)")

    Script.setUsageMessage(__doc__ + "\n".join(["Usage:", f"  {Script.scriptName} [option|cfgfile] ... "]))

    Script.parseCommandLine(ignoreErrors=True)

    chunk_size = 50000
    switches = Script.getUnprocessedSwitches()
    for switch in switches:
        if switch[0] == "ChunkSize":
            chunk_size = int(switch[1])

    executeSplitRunsForProd(dmScript.getBKQuery(), chunk_size)


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Takes an xmlSummary along with its IDs and creates a JSON file containing
all the counters in the xmlSummary."""
import json
import xml.etree.ElementTree as ET
from DIRAC import gLogger, S_ERROR, S_OK


def readXMLfile(xmlFile, jobID, prodID, wmsID, jsonFileName="errors_xmlSummary.json"):
    """The main execution function.

    :param str xmlFile: the name of the XML file
    :param str jobID: the job ID of the data
    :param str prodID: the production ID of the data
    :param str wmsID: the wms ID of the data
    """
    res = extractRoot(xmlFile)
    if not res["OK"]:
        return res
    root = res["Value"]
    res = extractData(root, jobID, prodID, wmsID)
    if not res["OK"]:
        return res
    jsonData = res["Value"]
    gLogger.notice(f"Attempting to read {xmlFile} and put save it as {jsonFileName}")
    createJSONfile(jsonData, jsonFileName)
    return S_OK()


##############################################


def extractRoot(xmlFile):
    """Takes the root element in the XML file.

    :param str xmlFile: the name of the XML file
    """
    try:
        gLogger.notice(f"Extracting the tree from the XML file {xmlFile}")
        tree = ET.parse(xmlFile)
    except Exception as inst:
        gLogger.error(f"ERROR: Cannot extract tree from XML file {xmlFile}")
        return S_ERROR(inst)
    return S_OK(tree.getroot())


##############################################


def extractData(root, jobID, prodID, wmsID):
    """Takes the root of the XML file and the ids of the job, and converts the
    json file from the XML file.

    :param str jobID: the job ID
    :param str prodID: the production ID
    :param str wmsID: the wms ID
    """
    gLogger.notice("Extracting data from the root of the XML file")
    idDict = {}
    idDict["JobID"] = jobID
    idDict["ProductionID"] = prodID
    idDict["wmsID"] = wmsID

    tempDict = {}
    result = {}

    if root.iter("counter") is None:
        return S_ERROR("ERROR: No counters in XML file!")
    for counter in root.iter("counter"):
        try:
            counterValue = int(counter.text)
        except ValueError as inst:
            return S_ERROR(str(inst))
        tempDict[counter.attrib["name"]] = counterValue

    # Checks if dict is empty (empty dicts evaluate to False)
    if not tempDict:
        gLogger.warn("WARNING: XML file is empty")

    result["Counters"] = tempDict
    result["Counters"]["ID"] = idDict
    return S_OK(json.dumps(result, indent=2))


##############################################


def createJSONfile(jsonData, jsonFileName):
    """Creates a JSON file given a string containing the data.

    :param str jsonData: the string that contains the data
    :param str jsonFileName: the name of the resulting JSON file
    """
    with open(jsonFileName, "w") as f:
        gLogger.notice(f"Writing the XML file to JSON file {jsonFileName}")
        f.write(jsonData)


# The file is run as following:
# readXMLfile(xmlFileName.xml, jobID, prodID, wmsID, jsonFileName.json)



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

import DIRAC
from DIRAC.Core.Base.Script import Script


def usage():
    from LHCbDIRAC.Interfaces.API.DiracProduction import DiracProduction

    print(f"Usage: {Script.scriptName} <Command> <Production ID> |<Production ID>")
    commands = DiracProduction().getProductionCommands()["Value"]
    print(f"\nCommands include: {', '.join(commands)}")
    print("\nDescription:\n")
    for n, v in commands.items():
        print(f"{n}:")
        for i, j in v.items():
            print(f"     {i} = {j}")

    DIRAC.exit(2)


@Script()
def main():
    Script.parseCommandLine(ignoreErrors=True)

    args = Script.getPositionalArgs()

    from LHCbDIRAC.Interfaces.API.DiracProduction import DiracProduction

    diracProd = DiracProduction()

    if len(args) < 2:
        usage()

    exitCode = 0
    errorList = []
    command = args[0]

    for prodID in args[1:]:
        result = diracProd.production(prodID, command, disableCheck=False)
        if "Message" in result:
            errorList.append((prodID, result["Message"]))
            exitCode = 2
        elif not result:
            errorList.append((prodID, "Null result for getProduction() call"))
            exitCode = 2
        else:
            exitCode = 0

    for error in errorList:
        print("ERROR %s: %s" % error)

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""reimplementation of the dictionary."""

from DIRAC import gLogger


def prepend(string, indent="___"):
    """add string."""
    string = string.strip("\n")
    tokens = string.split("\n")
    newstr = ""
    for token in tokens[0:-1]:
        newstr += indent + token + "\n"
    newstr += indent + tokens[-1]
    return newstr


class Entity(dict):
    """Entity class."""

    def __init__(self, properties={}):
        """initialize an Entity."""
        if isinstance(properties, list):
            for key in properties:
                self[key] = None  # find a simpler way to declare all keys
        elif isinstance(properties, dict):
            if properties:
                self.update(properties.items())
        else:
            gLogger.warn("Cannot create Entity from properties:" + str(properties))

    def __repr__(self):
        """print."""
        if not self:
            rString = "{\n " + str(None) + "\n}"
        else:
            rString = "{"
            keys = self.keys()
            if "fullpath" in keys:
                rString += "\n" + "fullpath: " + str(self["fullpath"])
            for key in keys:
                if key not in ("name", "level", "fullpath", "expandable", "selection", "method", "showFiles"):
                    rString += "\n " + str(key) + " : "
                    value = self[key]
                    # some entities do not have this key. Ignore then.
                    try:
                        if key in self["not2show"]:
                            rString += "-- not shown --"
                            continue
                    except Exception:
                        pass
                    if isinstance(value, dict):
                        value = Entity(value)
                        rString += "\n" + prepend(str(value), (len(str(key)) + 3) * " ")
                    else:
                        rString += str(value)
            else:
                for key in keys:
                    if key not in ("name", "fullpath", "FileName"):
                        value = self[key]
                        if isinstance(value, dict):
                            value = Entity(value)
                            rString += "\n" + prepend(str(value), (len(str(key)) + 3) * " ")
                        else:
                            rString += str(value)

            rString += "\n}"
        return rString



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""LHCbDIRAC.AccountingSystem.private.Plotters.PopularityPlotter.

PopularityPlotter.__bases__:
  DIRAC.AccountingSystem.private.Plotters.BaseReporter.BaseReporter
"""
from DIRAC import S_OK
from DIRAC.AccountingSystem.private.Plotters.BaseReporter import BaseReporter

from LHCbDIRAC.AccountingSystem.Client.Types.Popularity import Popularity

# FIXME: refactor _reportMethods
# FIXME: refactor _plotMethods


class PopularityPlotter(BaseReporter):
    """PopularityPlotter as extension of BaseReporter."""

    _typeName = "Popularity"
    _typeKeyFields = [dF[0] for dF in Popularity().definitionKeyFields]
    # FIXME: WTF is this ????, here includes StorageElement !!!
    _noSEtypeKeyFields = [dF[0] for dF in Popularity().definitionKeyFields]
    _noSEGrouping = (", ".join("%s" for f in _noSEtypeKeyFields), _noSEtypeKeyFields)

    # .............................................................................
    # data Usage

    _reportDataUsageName = "Data Usage"

    def _reportDataUsage(self, reportRequest):
        """Reports the data usage, from the Accounting DB.

        :param reportRequest: <dict>
          { 'groupingFields' : ( '%s', [ 'EventType' ] ),
            'startTime'      : 1355663249.0,
            'endTime'        : 1355749690.0,
            'condDict'       : { 'EventType' : '90000000' }
          }

        returns S_OK / S_ERROR
          { 'graphDataDict' : { '90000000' : { 1355616000L : 123.456789,
                                               1355702400L : 78.901234500000001 }
                                             },
            'data'          : { '90000000' : { 1355616000L : 123456.789,
                                               1355702400L : 78901.234500000006 }
                                             },
            'unit'          : 'kfiles',
            'granularity'   : 86400
          }
        """

        selectString = self._getSelectStringForGrouping(reportRequest["groupingFields"])
        selectFields = (
            selectString + ", %s, %s, SUM(%s)/SUM(%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", "Usage", "entriesInBucket"],
        )

        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            PopularityPlotter._noSEGrouping,
            {"convertToGranularity": "sum", "checkNone": True},
        )
        if not retVal["OK"]:
            return retVal

        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)

        accumMaxValue = self._getAccumulationMaxValue(dataDict)
        suitableUnits = self._findSuitableUnit(dataDict, accumMaxValue, "files")

        # 3rd value, maxValue is not used
        baseDataDict, graphDataDict, __, unitName = suitableUnits

        return S_OK(
            {"data": baseDataDict, "graphDataDict": graphDataDict, "granularity": granularity, "unit": unitName}
        )

    def _plotDataUsage(self, reportRequest, plotInfo, filename):
        """Creates <filename>.png file containing information regarding the data
        usage.

        :param reportRequest: <dict>
           { 'grouping'       : 'EventType',
             'groupingFields' : ( '%s', [ 'EventType' ] ),
             'startTime'      : 1355663249.0,
             'endTime'        : 1355749690.0,
             'condDict'       : { 'StorageElement' : 'CERN' }
           }
        :param plotInfo: <dict> ( output of _reportDataUsage )
           { 'graphDataDict' : { '90000001' : { 1355616000L : 223.45678899999999,
                                                1355702400L : 148.90123449999999 },
                                 '90000000' : { 1355616000L : 123.456789,
                                                1355702400L : 78.901234500000001 }
                                },
             'data'          : { '90000001' : { 1355616000L : 223456.78899999999,
                                                1355702400L : 148901.23449999999 },
                                 '90000000' : { 1355616000L : 123456.789,
                                                1355702400L : 78901.234500000006 }
                                },
             'unit'          : 'kfiles',
             'granularity': 86400
            }
        :param filename: <str>
          '_plotDataUsage'

        returns S_OK / S_ERROR
           { 'plot': True, 'thumbnail': False }
        """

        startEpoch = reportRequest["startTime"]
        endEpoch = reportRequest["endTime"]
        granularity = plotInfo["granularity"]
        dataDict = plotInfo["graphDataDict"]

        metadata = {
            "title": f"Data Usage grouped by {reportRequest['grouping']}",
            "starttime": startEpoch,
            "endtime": endEpoch,
            "span": granularity,
            "ylabel": plotInfo["unit"],
        }

        dataDict = self._fillWithZero(granularity, startEpoch, endEpoch, dataDict)
        return self._generateStackedLinePlot(filename, dataDict, metadata)

    # .............................................................................
    # normalized data Usage

    _reportNormalizedDataUsageName = "Normalized Data Usage"

    def _reportNormalizedDataUsage(self, reportRequest):
        """Reports the normalized data usage, from the Accounting DB.

        :param reportRequest: <dict>
          { 'groupingFields' : ( '%s', [ 'EventType' ] ),
            'startTime'      : 1355663249.0,
            'endTime'        : 1355749690.0,
            'condDict'       : { 'StorageElement' : 'CERN' }
          }

        returns S_OK / S_ERROR
          { 'graphDataDict' : { '90000001' : { 1355616000L : 223.45678899999999,
                                               1355702400L : 148.90123449999999 },
                                '90000000' : { 1355616000L : 123.456789,
                                               1355702400L : 78.901234500000001 }
                              },
            'data'          : { '90000001' : { 1355616000L : 223456.78899999999,
                                               1355702400L : 148901.23449999999 },
                                '90000000' : { 1355616000L : 123456.789,
                                               1355702400L : 78901.234500000006 }
                              },
            'unit'          : 'kfiles',
            'granularity'   : 86400
          }
        """

        selectString = self._getSelectStringForGrouping(reportRequest["groupingFields"])
        selectFields = (
            selectString + ", %s, %s, SUM(%s)/SUM(%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", "NormalizedUsage", "entriesInBucket"],
        )
        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            PopularityPlotter._noSEGrouping,
            {"convertToGranularity": "sum", "checkNone": True},
        )
        if not retVal["OK"]:
            return retVal

        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)

        accumMaxValue = self._getAccumulationMaxValue(dataDict)
        suitableUnits = self._findSuitableUnit(dataDict, accumMaxValue, "files")

        # 3rd value, maxValue is not used
        baseDataDict, graphDataDict, __, unitName = suitableUnits

        return S_OK(
            {"data": baseDataDict, "graphDataDict": graphDataDict, "granularity": granularity, "unit": unitName}
        )

    def _plotNormalizedDataUsage(self, reportRequest, plotInfo, filename):
        """Creates <filename>.png file containing information regarding the
        normalized data usage.

        :param reportRequest: <dict>
           { 'grouping'       : 'EventType',
             'groupingFields' : ( '%s', [ 'EventType' ] ),
             'startTime'      : 1355663249.0,
             'endTime'        : 1355749690.0,
             'condDict'       : { 'StorageElement' : 'CERN' }
           }
        :param plotInfo: <dict> ( output of _reportNormalizedDataUsage )
           { 'graphDataDict' : { '90000001' : { 1355616000L : 223.45678899999999,
                                                1355702400L : 148.90123449999999 },
                                 '90000000' : { 1355616000L : 123.456789,
                                                1355702400L : 78.901234500000001 }
                                },
             'data'          : { '90000001' : { 1355616000L : 223456.78899999999,
                                                1355702400L : 148901.23449999999 },
                                 '90000000' : { 1355616000L : 123456.789,
                                                1355702400L : 78901.234500000006 }
                                },
             'unit'          : 'kfiles',
             'granularity': 86400
            }
        :param filename: <str>
          '_plotNormalizedDataUsage'

        returns S_OK / S_ERROR
           { 'plot': True, 'thumbnail': False }
        """

        startEpoch = reportRequest["startTime"]
        endEpoch = reportRequest["endTime"]
        dataDict = plotInfo["graphDataDict"]
        granularity = plotInfo["granularity"]

        metadata = {
            "title": f"Normalized Data Usage grouped by {reportRequest['grouping']}",
            "starttime": startEpoch,
            "endtime": endEpoch,
            "span": granularity,
            "ylabel": plotInfo["unit"],
        }

        dataDict = self._fillWithZero(granularity, startEpoch, endEpoch, dataDict)
        return self._generateStackedLinePlot(filename, dataDict, metadata)



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Retrieve files for a given run."""
import DIRAC
from DIRAC.Core.Base.Script import Script


@Script()
def main():
    Script.setUsageMessage(
        __doc__
        + "\n".join(
            [
                "Usage:",
                f"  {Script.scriptName} [option|cfgfile] ... Run",
                "Arguments:",
                "  Run:      Run number (integer)",
            ]
        )
    )
    Script.parseCommandLine(ignoreErrors=True)
    args = Script.getPositionalArgs()

    try:
        runID = int(args[0])
    except (ValueError, IndexError):
        Script.showHelp(exitCode=1)

    exitCode = 0

    from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

    res = BookkeepingClient().getRunFiles(runID)
    if not res["OK"]:
        print(f"Failed to retrieve run files: {res['Message']}")
        exitCode = 2
    else:
        if not res["Value"]:
            print(f"No files found for run {runID}")
        else:
            print(f"{'FileName'.ljust(100)} {'Size'.ljust(10)} {'GUID'.ljust(40)} {'Replica'.ljust(8)}")
            for lfn in sorted(res["Value"]):
                size = res["Value"][lfn]["FileSize"]
                guid = res["Value"][lfn]["GUID"]
                hasReplica = res["Value"][lfn]["GotReplica"]
                print(f"{lfn.ljust(100)} {str(size).ljust(10)} {guid.ljust(40)} {str(hasReplica).ljust(8)}")

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""reads the data quality."""


class Quality:
    """Quality class."""

    def __init__(self):
        """initialize the class members."""
        self.group = ""
        self.flag = ""
        self.params = []

    def addParam(self, param):
        """adds a param."""
        self.params += [param]

    def __repr__(self):
        """formats the output of the print."""
        result = "Quality: "
        result += self.group + " " + self.flag + "\n"

        for param in self.params:
            result += str(param)

        result += "\n"
        return result

    def writeToXML(self):
        """creates an XML string."""
        return '<Quality Group="' + self.group + '" Flag="' + self.flag + '"/>\n'



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Unit tests for Workflow Modules."""
# pylint: disable=protected-access, missing-docstring, invalid-name, line-too-long

import unittest
import os
import copy
import shutil
from textwrap import dedent

from unittest.mock import MagicMock, patch

from DIRAC import gLogger
from DIRAC.TransformationSystem.Client.FileReport import FileReport

# mocks
from DIRAC.DataManagementSystem.Client.test.mock_DM import dm_mock
from DIRAC.Resources.Catalog.test.mock_FC import fc_mock


from LHCbDIRAC.Core.Utilities.XMLSummaries import XMLSummary

# mocks
from LHCbDIRAC.Workflow.Modules.mock_Commons import (
    step_commons,
    wf_commons,
    prod_id,
    prod_job_id,
    wms_job_id,
    workflowStatus,
    stepStatus,
    step_id,
    step_number,
    rc_mock,
)
from LHCbDIRAC.BookkeepingSystem.Client.test.mock_BookkeepingClient import bkc_mock

# sut
from LHCbDIRAC.Workflow.Modules.AnalyseXMLSummary import AnalyseXMLSummary


class ModulesTestCase(unittest.TestCase):
    """Base class for the Modules test cases."""

    def setUp(self):
        gLogger.setLevel("DEBUG")
        self.maxDiff = None

        self.jsu_mock = MagicMock()
        self.jsu_mock.setJobApplicationStatus.return_value = {"OK": True, "Value": ""}

        self.jsu_mock = MagicMock()
        self.jsu_mock.setJobApplicationStatus.return_value = {"OK": True, "Value": ""}

        self.ft_mock = MagicMock()
        self.ft_mock.transferAndRegisterFile.return_value = {"OK": True, "Value": {"uploadedSE": ""}}
        self.ft_mock.transferAndRegisterFileFailover.return_value = {"OK": True, "Value": {}}
        self.ft_mock.request = rc_mock
        self.ft_mock.FileCatalog = fc_mock

        self.nc_mock = MagicMock()
        self.nc_mock.sendMail.return_value = {"OK": True, "Value": ""}

        self.xf_o_mock = MagicMock()
        self.xf_o_mock.inputFileStats = {"a": 1, "b": 2}
        self.xf_o_mock.outputFileStats = {"a": 1, "b": 2}
        self.xf_o_mock.analyse.return_value = True

        self.jobStep_mock = MagicMock()
        self.jobStep_mock.commit.return_value = {"OK": True, "Value": ""}
        self.jobStep_mock.setValuesFromDict.return_value = {"OK": True, "Value": ""}
        self.jobStep_mock.checkValues.return_value = {"OK": True, "Value": ""}

    def tearDown(self):
        for fileProd in [
            "appLog",
            "foo.txt",
            "aaa.Bhadron.dst",
            "bbb.Calibration.dst",
            "bar.py",
            "aLongLog.log",
            "bookkeeping_123_00000456_321.xml",
            "aLongLog.log.gz",
            "ccc.charm.mdst",
            "ccc.charm.mdst",
            "prova.txt",
            "aLog.log",
            "BAR.txt",
            "FooBAR.ext.txt",
            "foo_1.txt",
            "bar_2.py",
            "bar.txt",
            "ErrorLogging_Step1_coredump.log",
            "123_00000456_request.xml",
            "lfn1",
            "lfn2",
            "XMLSummaryFile",
            "aaa.bhadron.dst",
            "bbb.calibration.dst",
            "ProductionOutputData",
            "data.py",
            "123_00000456_request.json",
            "00000123_00000456.tar",
            "someOtherDir",
            "DISABLE_WATCHDOG_CPU_WALLCLOCK_CHECK",
            "myfoo.blah",
            "prodConf_someApp__.py",
            "prodConf_someApp___.py",
        ]:
            try:
                os.remove(fileProd)
            except OSError:
                continue

        for directory in ["./job", "job"]:
            try:
                shutil.rmtree(directory)
            except BaseException:
                continue


@patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())
class AnalyseXMLSummarySuccess(ModulesTestCase):
    #################################################

    def test_execute(self, _patch):
        axlf = AnalyseXMLSummary(bkClient=bkc_mock, dm=dm_mock)
        axlf.stepInputData = ["some.sdst", "00012345_00006789_1.sdst"]
        axlf.jobType = "merge"

        logAnalyser = MagicMock()

        logAnalyser.return_value = True
        axlf.logAnalyser = logAnalyser
        axlf.XMLSummary_o = self.xf_o_mock
        axlf.nc = self.nc_mock
        axlf.XMLSummary = "XMLSummaryFile"
        with open(axlf.XMLSummary, "w") as f:
            f.write(
                dedent(
                    """<?xml version="1.0" encoding="UTF-8"?>

          <summary version="1.0" xsi:noNamespaceSchemaLocation="$XMLSUMMARYBASEROOT/xml/XMLSummary.xsd" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
                  <success>True</success>
                  <step>finalize</step>
                  <usage>
                          <stat unit="KB" useOf="MemoryMaximum">866104.0</stat>
                  </usage>
                  <input>
                          <file GUID="CCE96707-4BE9-E011-81CD-003048F35252" name="LFN:00012478_00000532_1.sim" status="full">200</file>
                  </input>
                  <output>
                          <file GUID="229BBEF1-66E9-E011-BBD0-003048F35252" name="PFN:00012478_00000532_2.xdigi" status="full">200</file>
                  </output>
          </summary>
          """  # noqa
                )
            )

        # no errors, all ok
        for wf_cs in copy.deepcopy(wf_commons):
            for s_cs in step_commons:
                self.assertTrue(
                    axlf.execute(
                        prod_id, prod_job_id, wms_job_id, workflowStatus, stepStatus, wf_cs, s_cs, step_number, step_id
                    )["OK"]
                )

        # logAnalyser gives errors
        axlf.jobType = "reco"

        logAnalyser.return_value = False
        axlf.logAnalyser = logAnalyser

        for wf_cs in copy.deepcopy(wf_commons):
            for s_cs in step_commons:
                self.assertTrue(
                    axlf.execute(
                        prod_id, prod_job_id, wms_job_id, workflowStatus, stepStatus, wf_cs, s_cs, step_number, step_id
                    )["OK"]
                )

    def test__basicSuccess(self, _patch):
        axlf = AnalyseXMLSummary(bkClient=bkc_mock, dm=dm_mock)

        with open("XMLSummaryFile", "w") as f:
            f.write(
                dedent(
                    """<?xml version="1.0" encoding="UTF-8"?>

          <summary version="1.0" xsi:noNamespaceSchemaLocation="$XMLSUMMARYBASEROOT/xml/XMLSummary.xsd" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
                  <success>True</success>
                  <step>finalize</step>
                  <usage>
                          <stat unit="KB" useOf="MemoryMaximum">866104.0</stat>
                  </usage>
                  <input>
                          <file GUID="CCE96707-4BE9-E011-81CD-003048F35252" name="LFN:00012478_00000532_1.sim" status="full">200</file>
                  </input>
                  <output>
                          <file GUID="229BBEF1-66E9-E011-BBD0-003048F35252" name="PFN:00012478_00000532_2.xdigi" status="full">200</file>
                  </output>
          </summary>
          """  # noqa
                )
            )
        axlf.XMLSummary_o = XMLSummary("XMLSummaryFile")
        res = axlf._basicSuccess()
        self.assertFalse(res)

        axlf.XMLSummary_o.inputFileStats = {"full": 2, "part": 1, "fail": 0, "other": 0}
        axlf.XMLSummary_o.inputStatus = [("aa/1.txt", "full"), ("aa/2.txt", "part")]
        axlf.inputDataList = ["aa/1.txt", "aa/2.txt"]
        axlf.numberOfEvents = -1
        axlf.fileReport = FileReport()
        axlf.production_id = "123"
        res = axlf._basicSuccess()
        self.assertTrue(res)
        self.assertEqual(axlf.fileReport.statusDict, {"aa/2.txt": "Problematic"})

        axlf.XMLSummary_o.inputFileStats = {"full": 2, "part": 0, "fail": 1, "other": 0}
        axlf.XMLSummary_o.inputStatus = [("aa/1.txt", "fail"), ("aa/2.txt", "full")]
        axlf.inputDataList = ["aa/1.txt", "aa/2.txt"]
        axlf.numberOfEvents = -1
        axlf.fileReport = FileReport()
        axlf.production_id = "123"
        res = axlf._basicSuccess()
        self.assertTrue(res)
        self.assertEqual(axlf.fileReport.statusDict, {"aa/1.txt": "Problematic"})

        axlf.XMLSummary_o.inputFileStats = {"full": 2, "part": 0, "fail": 1, "other": 0}
        axlf.XMLSummary_o.inputStatus = [("aa/1.txt", "fail"), ("aa/2.txt", "full")]
        axlf.inputDataList = ["aa/3.txt"]
        axlf.numberOfEvents = -1
        axlf.fileReport = FileReport()
        axlf.production_id = "123"
        res = axlf._basicSuccess()
        self.assertFalse(res)
        self.assertEqual(axlf.fileReport.statusDict, {})

        axlf.XMLSummary_o.inputFileStats = {"full": 2, "part": 1, "fail": 1, "other": 0}
        axlf.XMLSummary_o.inputStatus = [("aa/1.txt", "fail"), ("aa/2.txt", "part")]
        axlf.inputDataList = ["aa/1.txt", "aa/2.txt"]
        axlf.numberOfEvents = -1
        axlf.fileReport = FileReport()
        axlf.production_id = "123"
        res = axlf._basicSuccess()
        self.assertTrue(res)
        self.assertEqual(axlf.fileReport.statusDict, {"aa/1.txt": "Problematic", "aa/2.txt": "Problematic"})

        axlf.XMLSummary_o.inputFileStats = {"full": 2, "part": 1, "fail": 1, "other": 0}
        axlf.XMLSummary_o.inputStatus = [("aa/1.txt", "fail"), ("aa/2.txt", "part")]
        axlf.inputDataList = ["aa/1.txt", "aa/2.txt"]
        axlf.numberOfEvents = "10"
        axlf.fileReport = FileReport()
        axlf.production_id = "123"
        res = axlf._basicSuccess()
        self.assertTrue(res)
        self.assertEqual(axlf.fileReport.statusDict, {"aa/1.txt": "Problematic"})


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Retrieve information from the Bookkeeping for a given production."""
import DIRAC
from DIRAC.Core.Base.Script import Script


@Script()
def main():
    Script.setUsageMessage(
        __doc__
        + "\n".join(
            [
                "Usage:",
                f"  {Script.scriptName} [option|cfgfile] ... ProdID",
                "Arguments:",
                "  ProdID:   Production ID (mandatory)",
            ]
        )
    )
    Script.parseCommandLine(ignoreErrors=True)
    args = Script.getPositionalArgs()

    if len(args) < 1:
        Script.showHelp(exitCode=1)

    from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

    bk = BookkeepingClient()
    prod = int(args[0])

    res = bk.getProductionInformation(prod)
    if not res["OK"]:
        print(f"ERROR {prod}: {res['Message']}")
        DIRAC.exit(2)

    val = res["Value"]
    print("Production Info: ")

    infs = val["Production information"]
    if infs is not None:
        for inf in infs:
            if inf[2] is not None:
                print("    Configuration Name:", inf[0])
                print("    Configuration Version:", inf[1])
                print("    Event type:", inf[2])

    steps = val["Steps"]

    for step in steps:
        print("-----------------------")
        print(f" StepName: {step[0]} ")
        print(f"    StepId             : {step[7]}")
        print(f"    ApplicationName    : {step[1]}")
        print(f"    ApplicationVersion : {step[2]}")
        print(f"    OptionFiles        : {step[3]}")
        print(f"    DDB                : {step[4]}")
        print(f"    CONDDB             : {step[5]}")
        print(f"    ExtraPackages      : {step[6]}")
        print(f"    Visible            : {step[8]}")
        print("-----------------------")
    print("Number of Steps  ", val["Number of jobs"][0][0])
    files = val["Number of files"]
    if files:
        print("Total number of files:", files[0][2])
    else:
        print("Total number of files: 0")
    for file in files:
        print("         " + str(file[1]) + ":" + str(file[0]))
    nbevent = val["Number of events"]
    if nbevent:
        print("Number of events")
        print("File Type".ljust(20) + "Number of events".ljust(20) + "Event Type".ljust(20) + "EventInputStat")
        for i in nbevent:
            print(str(i[0]).ljust(20) + str(i[1]).ljust(20) + str(i[2]).ljust(20) + str(i[3]))
    else:
        print("Number of events", 0)
    print("Path: ", val["Path"])


if __name__ == "__main__":
    main()
